#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup
    from Owlcat.Plotting import MultigridPlot, SkyPlot, PLOT_SINGLE, PLOT_MULTI, PLOT_ERRORBARS

    parser = OptionParser(usage="""%prog: [plots & options] <parmtable(s) or cache file>""",
                          description="""Makes various plots of dE solutions.""")

    parser.add_option("-l", "--list", action="count",
                      help="lists stuff found in MEP tables, then exits. Use multiple times for more detail.")
    parser.add_option("--unnormalized", action="store_true",
                      help="dE amplitude is usually normalized across all antennas, to take out LSM effects. Use this option to disable this.")
    parser.add_option("--ampl", action="store_true",
                      help="makes amplitude plots")
    parser.add_option("--phase", action="store_true",
                      help="makes phase plots")
    parser.add_option("--pol", action="store_true",
                      help="makes instrumental polarization plots")
    parser.add_option("--per-corr", action="store_true",
                      help="includes per-correlation plots in the above")
    parser.add_option("--per-band", action="store_true",
                      help="includes combined per-band plots in the above")
    parser.add_option("--circle-ampl", action="store_true",
                      help="makes circle plots of average dE-amplitudes per antenna")
    parser.add_option("--circle-phase", action="store_true",
                      help="makes circle plots of average dE-phases per antenna")
    parser.add_option("--circle-ampl-ant", metavar="ANTENNA", type="string",
                      help="makes per-timeslot plots of average dE-amplitudes for the given antenna")
    parser.add_option("--circle-phase-ant", metavar="ANTENNA", type="string",
                      help="makes per-timeslot plots of average dE-phases for the given antenna")
    parser.add_option("--ampl-slope", action="store_true",
                      help="makes plots of an East-West slope fit for dE-amplitudes")
    parser.add_option("--phase-slope", action="store_true",
                      help="makes plots of a East-West slope fit for dE-phases")
    parser.add_option("--phase-slope-rot-offset", metavar="DEG", type="float",
                      help="...including phase slopes corresponding to rotational offset (EXPERIMENTAL)")
    parser.add_option("--phase-slope-time-offset", metavar="FRAC", type="float",
                      help="...including phase slopes corresponding to time offset (EXPERIMENTAL)")
    parser.add_option("--phase-slope-res", action="store_true",
                      help="makes plots of residuals w.r.t. slope fit for dE-phases")
    parser.add_option("--phase-slope-dlm", action="store_true",
                      help="makes plots of dl/dm offsets corresponding to phase slopes (EXPERIMENTAL)")
    parser.add_option("--lsm", type="string",
                      help="LSM file (for source positions where needed). Default is first *.lsm.html file found.")
    parser.add_option("--update-lsm", action="store_true",
                      help="updates LSM file based on dE solutions (EXPERIMENTAL)")
    parser.add_option("--ms", type="string",
                      help="Measurement Set (for UVW coordinates, antenna positions and such)")
    parser.add_option("--antennas", type="string",
                      help="antenna subset, as a comma-separated list of names. Simple wildcards are allowed.")
    parser.add_option("--sources", type="string",
                      help="source subset, as a comma-separated list of names. Simple wildcards are allowed. Use -l/--list to list all antennas and sources defined in the given table(s).")
    parser.add_option("-c", "--cache", metavar="FILENAME", type="string",
                      help="cache parms to file, which can be given instead of parmtables next time, for quicker startup")
    parser.add_option("--pe", metavar="FILENAME", type="string",
                      help="cache file from plot-pointing-errors.py script. If given, |dE| given by the pointing model will be plotted")

    plotgroup = OptionGroup(parser, "Plotting options")
    outputgroup = OptionGroup(parser, "Output options")
    MultigridPlot.init_options(plotgroup, outputgroup)
    SkyPlot.init_options(plotgroup, outputgroup)

    parser.add_option_group(plotgroup)
    parser.add_option_group(outputgroup)

    parser.set_defaults(circle_ampl_ant="", circle_phase_ant="", antennas="", sources="", cache=None,
                        borders="0.05,0.99,0.05,0.95", circle_borders=".05,.9,.05,.9", subplot_wspace=0,
                        title_fontsize=12, subtitle_fontsize=10, label_fontsize=8, axis_fontsize=5,
                        circle_title_fontsize=12, circle_label_fontsize=8, circle_axis_fontsize=5,
                        phase_slope_rot_offset=0,
                        radius=0, dlm_scale=0,
                        circle_maxsize=36, circle_minsize=0, circle_linewidth=5, circle_label_offset=0.02,
                        output_prefix="", output_type="png", papertype='a4', width=0, height=0)

    (options, args) = parser.parse_args()

    if not options.circle_ampl and not options.circle_phase and \
            not options.circle_ampl_ant and not options.circle_phase_ant and \
            not options.ampl_slope and not options.phase_slope and not options.phase_slope_res and \
            not options.ampl and not options.phase and not options.pol and \
            not options.list:
        parser.error("No plots specified. Use -h for help.")

    if not args:
        parser.error("No parmtables or cachefile specified. Use -h foe help.")

    if (options.width or options.height) and not (options.width and options.height):
        parser.error("Either both -W/--width and -H/--height must be specified, or neither.")

    try:
        borders = [float(x) for x in options.borders.split(",")]
    except:
        borders = []
    if len(borders) != 4:
        parser.error("Bad --borders specification.")
    try:
        circle_borders = [float(x) for x in options.circle_borders.split(",")]
    except:
        circle_borders = []
    if len(circle_borders) != 4:
        parser.error("Bad --circle-borders specification.")

    if not options.grid_circle:
        options.grid_circle = (30, 60)

    import os.path
    import os
    import pickle
    import sys
    # load other stuff
    import numpy
    from scipy import linalg
    from Owlcat import Coordinates

    import math

    DEG = math.pi / 180
    ARCMIN = math.pi / (180 * 60)

    # === loads MS (that has been specified in options, else picks one in current dir)
    # === returns tuple of ms,msant, where the first is the main table, and the second is the ANTENNA subtable
    _ms = _msant = None


    def load_ms():
        global _ms
        global _msant
        if _ms is None:
            import pyrap.tables
            # if MS file not specified, use first one found
            if not options.ms:
                mss = [filename for filename in os.listdir(".") if filename.lower().endswith(".ms")]
                if not mss:
                    parser.error("No MSs found. Use --ms to specify one explicitly.")
                msname = mss[0]
            else:
                msname = options.ms
            print(("Using MS %s for coordinates information" % msname))
            _ms = pyrap.tables.table(msname)
            _msant = pyrap.tables.table(_ms.getkeyword('ANTENNA'))
        return _ms, _msant


    #
    # ========================== LOAD LSM FILE AND MS
    #
    if options.circle_ampl or options.circle_phase or options.circle_ampl_ant or options.circle_phase_ant \
            or options.phase_slope_rot_offset or options.phase_slope_time_offset or options.phase_slope_dlm \
            or options.update_lsm or (options.pe and options.ampl):
        # if LSM file not specified, use first one found
        if not options.lsm:
            lsms = [filename for filename in os.listdir(".") if filename.endswith(".lsm.html")]
            if not lsms:
                parser.error("No LSMs found. Use --lsm to specify one explicitly.")
            lsm_filename = lsms[0]
        else:
            lsm_filename = options.lsm
        #
        # find tigger
        try:
            import Tigger
        except ImportError:
            # make plot of average dE to model
            sys.path.append(os.getenv('HOME'))
            import Tigger
        print(("Using LSM file %s" % lsm_filename))
        model = Tigger.load(lsm_filename)
        lsm_timestamp = os.path.getmtime(lsm_filename)

        # if MS file not specified, use first one found
        ms, msant = load_ms()
        nant = msant.nrows()
        # read UVWs
        uvw0 = ms.query('ANTENNA1==0 && ANTENNA2==%d' % (nant - 1)).getcol('UVW')
        uvw = uvw0[30::60]
        # read phase center
        import pyrap.tables

        radec0 = pyrap.tables.table(ms.getkeyword('FIELD')).getcol('PHASE_DIR')[0][0]
        print(("Phase center is at", radec0))
    else:
        model = None

    #
    # ========================== read parmtables
    #
    mep_timestamp = os.path.getmtime(args[0])

    PREFIX = 'dE'


    def make_funklet_name(*elements):
        return ':'.join([PREFIX] + list(elements))


    if len(args) > 1 or not args[0].endswith('.cache'):
        from Owlcat.ParmTables import ParmTab

        # get antenna positions
        if options.ampl_slope or options.phase_slope or \
                options.phase_slope_rot_offset or options.phase_slope_time_offset or options.phase_slope_dlm:
            ms, msant = load_ms()
            names = list(msant.getcol('NAME'))
            pos = msant.getcol('POSITION')
            # get reference position
            x, y, z = pos0 = pos[0, :]
            # convert to unit vector pointing east
            dir_east = numpy.array([-y, x, 0])
            dir_east /= math.sqrt((dir_east ** 2).sum())
            # get coordinate of each antenna along dir_east: this is just the scalar product of
            # (pos[i,:]-pos[0,:]) by dir_east.
            antx = ((pos - pos0[numpy.newaxis, :]) * dir_east[numpy.newaxis, :]).sum(1)
            # now convert to dict of antenna coordinates
            ANTX_dict = dict(list(zip(names, antx)))
            # update with abbreviated antenna names, if they share a common prefix
            prefix = 1
            while prefix < len(names[0]) and all([n[:prefix] == names[0][:prefix] for n in names[1:]]):
                prefix += 1
            if prefix > 1:
                ANTX_dict.update([(n[prefix - 1:], ax) for n, ax in zip(names, antx)])
            # print "Antenna positions: "," ".join([ "%s:%f"%(n,ANTX_dict[n]) for n in sorted(ANTX_dict.keys()) ])
        # else antenna positions not needed
        else:
            ANTX_dict = ANTX = None

        # spectral windows
        SPWS = list(range(len(args)))
        NCHANS = [0] * len(SPWS)
        # set of all sources, antennas and correlations
        SRCS = set()
        ANTS = set()
        CORRS = set()

        # complex array of dEs per each src,antenna,corr tuple
        des = {}

        # vector of frequencies
        freq0 = numpy.array([0.] * len(SPWS))

        # scan funklet names to build up sets of keys
        pt = ParmTab(args[0])
        for name in pt.funklet_names():
            label, src, ant, corr, ri = name.split(':')
            SRCS.add(src)
            ANTS.add(ant)
            CORRS.add(corr)

        # see how many timeslots we have per each combination of source, antenna, correlation
        funklet_counts = {}

        SRCS = sorted(SRCS)
        ANTS = sorted(ANTS)
        CORRS = sorted(CORRS)
        for i, src in enumerate(SRCS):
            for j, ant in enumerate(ANTS):
                for k, corr in enumerate(CORRS):
                    arr = pt.funkset(make_funklet_name(src, ant, corr, "r")).get_slice().array()
                    ntime, nfreq = arr.shape if arr.ndim == 2 else (len(arr), 1)
                    funklet_counts[src, ant, corr] = ntime, nfreq

        if not funklet_counts:
            print(("No dE solutions found in MEP table %s" % args[0]))
            sys.exit(1)

        NTIMES = max([fc[0] for fc in list(funklet_counts.values())])
        NFREQS = max([fc[1] for fc in list(funklet_counts.values())])

        if options.list:
            print(("MEP table %s contains dE's for:" % args[0]))
            print(("%d correlations: %s" % (len(CORRS), " ".join(CORRS))))
            print(("%d antennas: %s" % (len(ANTS), " ".join(ANTS))))
            print(("%d sources: %s" % (len(SRCS), " ".join(["%d:%s" % (i, src) for i, src in enumerate(SRCS)]))))
            print(("%d timeslots, %d frequencies" % (NTIMES, NFREQS)))

            if options.list > 1:
                print("Per-source, per-antenna dE solution counts:")
                for i, src in enumerate(SRCS):
                    for j, ant in enumerate(ANTS):
                        for k, corr in enumerate(CORRS):
                            rr = pt.funkset(make_funklet_name(src, ant, corr, "r")).get_slice().array()
                            ii = pt.funkset(make_funklet_name(src, ant, corr, "i")).get_slice().array()
                            print((
                                "  %s: %s real, %s imaginary" % (make_funklet_name(src, ant, corr), rr.shape, ii.shape)))

            sys.exit(0)

        # source subset selection
        import fnmatch

        if options.sources:
            srcs = set()
            for ss in options.sources.split(","):
                subset = fnmatch.filter(SRCS, ss)
                if subset:
                    srcs.update(subset)
                else:
                    print(("WARNING: \"%s\" does not match any source names in MEP table %s." % (ss, args[0])))
                srcs.update(fnmatch.filter(SRCS, ss))
            SRCS = sorted(srcs)
            if not SRCS:
                print("No sources were selected, check your --sources option.")
                sys.exit(1)
        print(("Using %d sources: %s" % (len(SRCS), " ".join(SRCS))))

        # antenna subset selection
        if options.antennas:
            ants = set()
            for a in options.antennas.split(","):
                subset = fnmatch.filter(ANTS, a)
                if subset:
                    ants.update(subset)
                else:
                    print(("WARNING: \"%s\" does not match any antenna names in MEP table %s." % (a, args[0])))
            ANTS = sorted(ants)
            if not ANTS:
                print("No antennas were selected, check your --antennas option.")
                sys.exit(1)
        print(("Using %d antennas: %s" % (len(ANTS), " ".join(ANTS))))

        # check that antenna positions are known
        if ANTX_dict:
            unknown_antennas = [p for p in ANTS if p not in ANTX_dict]
            if unknown_antennas:
                print(("Don't have positions for antenna(s) %s" % ",".join(unknown_antennas)))
                print("Perhaps you should specify a proper MS with --ms?")
                sys.exit(1)

            # make vector of antenna positions
            ANTX = numpy.array([ANTX_dict[p] for p in ANTS])
            # recenter at middle of the array
            ANTX -= ANTX[-1] / 2


        def c00(funklet):
            if numpy.isscalar(funklet.coeff):
                return funklet.coeff
            else:
                return funklet.coeff.ravel()[0]


        def read_complex_parms(pt, freq0, prefix="dE"):
            """Reads complex c00 funklets from the given parmtable (given as filename).
            For each key in the 'keys' list, forms up keystring (by joining the elements
            of key weith ":"), then reads <prefix>:<keystring>:r and reads <prefix>:<keystring>:i,
            and forms up and returns a complex array.
            This assumes that the parm only changes in time.
            """
            # mask = numpy.zeros((len(SRCS),len(ANTS),len(CORRS),NTIMES),dtype=bool)
            arr = numpy.zeros((len(SRCS), len(ANTS), len(CORRS), NTIMES, NFREQS), dtype=complex)
            # arr = numpy.ma.masked_array(mask,arr,fill_value=1.0)
            freq_filled = False
            for i, src in enumerate(SRCS):
                for j, ant in enumerate(ANTS):
                    for k, corr in enumerate(CORRS):
                        fsr = pt.funkset(make_funklet_name(src, ant, corr, "r"))
                        fsi = pt.funkset(make_funklet_name(src, ant, corr, "i"))
                        fs_real = fsr.get_slice().array()
                        fs_imag = fsi.get_slice().array()
                        if not freq_filled:
                            for ifreq, funk in enumerate(fsr.get_slice(time=0)):
                                freq0[ifreq] = sum(funk.domain.freq) / 2
                            freq_filled = True
                        # init array with ones (if we have a shorter number of solutions, then we fill only the beginning)
                        arr[i, j, k, :, :] = 1
                        shape = [min(i1, i2) for i1, i2 in zip(fs_real.shape, fs_imag.shape)]
                        shape[1] = min(shape[1], NFREQS)
                        if shape[0] and shape[1]:
                            arr[i, j, k, :shape[0], :shape[1]].real = fs_real[:shape[0], :shape[1]]
                            arr[i, j, k, :shape[0], :shape[1]].imag = fs_imag[:shape[0], :shape[1]]
            return arr


        # stuff down the line does not handle masked arrays properly, so we don't use them
        # mask = numpy.zeros((len(SPWS),len(SRCS),len(ANTS)+1,len(CORRS),NTIMES),dtype=bool)
        de = numpy.zeros((len(SPWS) * NFREQS, len(SRCS), len(ANTS) + 1, len(CORRS), NTIMES), dtype=complex)
        freq0 = numpy.zeros(len(SPWS) * NFREQS, float)
        # de = numpy.ma.masked_array(de,mask,fill_value=1.0)

        for spw, tabname in enumerate(args):
            print(("Reading", tabname))
            pt = ParmTab(tabname)
            mep_timestamp = max(mep_timestamp, os.path.getmtime(tabname))
            fq0 = numpy.zeros(NFREQS, float)
            de1 = read_complex_parms(pt, fq0)
            de[spw * NFREQS:(spw + 1) * NFREQS, :, 0:len(ANTS), :, :] = read_complex_parms(pt, fq0).transpose(
                (4, 0, 1, 2, 3))
            freq0[spw * NFREQS:(spw + 1) * NFREQS] = fq0

        #    print de[0,0,0,0,:]
        SPWS = list(range(len(SPWS) * NFREQS))
        print(("Read %d parmtables, %d timeslots, %d frequencies" % (len(args), NTIMES, len(SPWS))))
        print(("Frequencies are %s MHz" % ",".join(["%d" % round(f * 1e-6) for f in freq0])))

        # write cache
        if options.cache:
            cachefile = options.cache + '.cache'
            pickle.dump((freq0, de, SPWS, SRCS, ANTS, CORRS, NTIMES, ANTX), open(cachefile, 'w'))
            print(("Cached all structures to file", cachefile))

    #
    # ========================== read cache file
    #
    else:
        print(("Reading cache file", args[0]))
        freq0, de, SPWS, SRCS, ANTS, CORRS, NTIMES, ANTX = pickle.load(open(args[0]))
        print(("Read %s: %d spws, %d srcs, %d ants, %d corrs, %d times" % (args[0],
                                                                          len(SPWS), len(SRCS), len(ANTS), len(CORRS),
                                                                          NTIMES)))

    # check options that specify antennas by name, to make sure antenna is known
    for ant in 'circle_ampl_ant', 'circle_phase_ant':
        antname = getattr(options, ant, None)
        if antname:
            try:
                globals()[ant] = ANTS.index(antname)
            except IndexError:
                parser.error("Antenna name '%s' not found." % antname)
        else:
            globals()[ant] = None

    # read pointing errors cache
    if options.pe:
        pe_dlm, pe_bsz, pe_beam_sizes = pickle.load(open(options.pe,'rb'))

        # pe_dlm is 2 x NSPW x NANT x NTIME
        # pe_bsz is 2 x 2 x NSPW x NANT x NTIME
        if pe_dlm.shape != (2, len(SPWS), len(ANTS), NTIMES) or \
                pe_bsz.shape != (2, 2, len(SPWS), len(ANTS), NTIMES):
            print(("Shape of cached arrays in file %s does not match: %s, %s" % (options.pe, pe_dlm.shape, pe_bsz_shape)))
            sys.exit(1)
        print(("Read pointing errors and %d beam extent(s) from %s" % (pe_beam_sizes, options.pe)))
        if pe_beam_sizes > 1:
            print("WARNING: more than 1 beam extent is not currently supported")

    # de is an NSPW x NSRC x NANT x NCORR x NTIME array of complex dEs
    # (NANT+1 actually, the last one is the mean-across-antennas value)
    #

    # some more constants
    MEAN = len(ANTS);  # index of "mean" antenna
    for C1, C2 in ('xx', 'yy'), ('XX', 'YY'), ('rr', 'll'), ('RR', 'LL'):
        if C1 in CORRS and C2 in CORRS:
            print(("Using %s/%s Jones elements" % (C1, C2)))
            XX = CORRS.index(C1)
            YY = CORRS.index(C2)
            break
    else:
        if len(CORRS) == 1:
            XX = YY = 0
            print(("Single-polarization data, using %s Jones element" % CORRS[0]))
        else:
            print("Can't find xx/yy or rr/ll correlations in MEP table")
            sys.exit(1)

    # set of all source,antenna,corr combinations
    ALL = [(src, ant, corr) for src in SRCS for ant in ANTS for corr in CORRS]

    # print per-source mean de_xx and de_yy amplitudes
    # (skip the last antenna since it's the MEAN one.)
    dex = abs(de[:, :, :-1, XX, :]).mean(2).mean(2).mean(0)
    dey = abs(de[:, :, :-1, YY, :]).mean(2).mean(2).mean(0)
    print("=== Mean |dExx|, |dEyy|:")
    for isrc, src in enumerate(SRCS):
        print(("%8s=[%16.8f,%16.8f]," % (src, dex[isrc], dey[isrc])))

    # take mean along antenna axis
    de[:, :, MEAN, :, :] = de1 = numpy.mean(de[:, :, 0:len(ANTS), :, :], 2)

    # get phase component
    de_phase_rad = numpy.angle(de)
    de_phase = de_phase_rad * 180 / math.pi
    de_ampl = abs(de)
    # average XX/YY dE phase (NSPWxNSRCxNANTxNTIME)
    deph = de_phase_rad.mean(3)
    # average spw/XX/YY dE phase (NSRCxNANTxNTIME)
    deph_mean = deph.mean(0)

    ### phase analysis
    # subtract mean phase over all antennas and times
    # dep0 is what's left, NSPW x NSRC x NANT x NCORR x NTIME
    dep0 = de_phase - (de_phase.mean(4).mean(2))[:, :, numpy.newaxis, :, numpy.newaxis]
    # Then we take the mean across correlations. dep0 becomes NSPW x NSRC x NANT x NTIME.
    dep0 = dep0.mean(3)
    # cut out the last ("mean") antenna
    dep0 = dep0[:, :, 0:-1, :]

    if ANTX is not None:
        # now fit phase offset and slope over array
        # dep0_b0/1 will be NSPWxNSRCxNCORRxNTIME
        sx = ANTX.sum()
        sx2 = (ANTX ** 2).sum()
        sy = dep0.sum(2)
        sxy = (dep0 * ANTX[numpy.newaxis, numpy.newaxis, :, numpy.newaxis]).sum(2)
        n = len(ANTX)
        ## slope of linear fit
        dep0_b1 = (n * sxy - sy * sx) / (n * sx2 - sx ** 2)
        ## offset of linear fit
        dep0_b0 = sy / n - dep0_b1 * (sx / n)
        ## stddev w.r.t. linear fit
        # deph_fit_std = (deph - deph_b1*fq[:,numpy.newaxis,numpy.newaxis,numpy.newaxis] - deph_b0).std(0)

        # now also compute residuals w.r.t. the slope fits, NSPW x NSRC x NANT x NTIME
        dep0_fitted = dep0_b0[:, :, numpy.newaxis, :] + \
                      dep0_b1[:, :, numpy.newaxis, :] * ANTX[numpy.newaxis, numpy.newaxis, :, numpy.newaxis]
        dep0_res = dep0 - dep0_fitted

        # same analysis for amplitudes
        dea0 = de_ampl - (de_ampl.mean(4).mean(2))[:, :, numpy.newaxis, :, numpy.newaxis]
        # cut out the last ("mean") antenna
        dea0 = dea0[:, :, 0:-1, :, :]

        # now fit phase offset and slope over array
        # dep0_b0/1 will be NSPWxNSRCxNCORRxNTIME
        sy = dea0.sum(2)
        sxy = (dea0 * ANTX[numpy.newaxis, numpy.newaxis, :, numpy.newaxis, numpy.newaxis]).sum(2)
        n = len(ANTX)
        ## slope of linear fit
        dea0_b1 = (n * sxy - sy * sx) / (n * sx2 - sx ** 2)
        ## offset of linear fit
        dea0_b0 = sy / n - dea0_b1 * (sx / n)
        ## stddev w.r.t. linear fit
        # deph_fit_std = (deph - deph_b1*fq[:,numpy.newaxis,numpy.newaxis,numpy.newaxis] - deph_b0).std(0)

        # now also compute residuals w.r.t. the slope fits
        dea0_fitted = dea0_b0[:, :, numpy.newaxis, :, :] + \
                      dea0_b1[:, :, numpy.newaxis, :, :] * ANTX[numpy.newaxis, numpy.newaxis, :, numpy.newaxis,
                                                           numpy.newaxis]
        dea0_res = dea0 - dea0_fitted

    # unnormalized average dE (for pointing error plots), NSPWxNSRCxNANTxNTIME
    de_unnorm = numpy.sqrt((abs(de[:, :, :, XX, :]) ** 2 + abs(de[:, :, :, YY, :]) ** 2) / 2)
    de_unnorm[:, :, MEAN, :] = numpy.mean(de_unnorm[:, :, 0:len(ANTS), :], 2)

    # now normalize each dE by mean dE across all times and antennas (this takes out
    # corrections due to imperfect LSMs)
    if not options.unnormalized:
        de /= numpy.mean(de1, 3)[:, :, numpy.newaxis, :, numpy.newaxis]

    # add "mean" to antenna labels
    ANTS += ["mean"]

    # phase
    # de_phase = numpy.angle(de)*180/math.pi

    # de_norm[spw,src,ant,time] is sqrt(|dExx|^2+|dEyy|^2), this is the "I" gain
    de_norm = numpy.sqrt((abs(de[:, :, :, XX, :]) ** 2 + abs(de[:, :, :, YY, :]) ** 2) / 2)
    de_norm[:, :, MEAN, :] = numpy.mean(de_norm[:, :, 0:len(ANTS), :], 2)
    # de_pol[spw,src,ant,time] is sqrt(|dExx|^2-|dEyy|^2), this is the instrumental Q polarization
    # (subtracted mean over antennas and times to take out LSM bias)
    de_pol = (abs(de[:, :, :, XX, :]) ** 2 - abs(de[:, :, :, YY, :]) ** 2) / 2
    de_pol[:, :, MEAN, :] = numpy.mean(de_pol[:, :, 0:len(ANTS), :], 2)
    de_pol -= numpy.mean(de_pol[:, :, MEAN, :], 2)[:, :, numpy.newaxis, numpy.newaxis]
    # de_fpol[spw,src,ant,time] is de_norm/de_polz, i.e. fractional instrumental Q
    de_fpol = de_pol / (de_norm ** 2)
    # de_renorm[spw,src,ant,time] is de_norm divided by the mean de_norm[spw,src,:,time],
    # i.e. mean gain across all antennas
    if options.unnormalized:
        de_renorm = de_norm
    else:
        de_renorm = de_norm / ((de_norm[:, :, MEAN, :])[:, :, numpy.newaxis, :])

    # read in source positions, if plotting mode requires it
    if model:
        maxabsdelog = math.sqrt(0.5);  # abs(delog).max()
        maxsize = options.circle_maxsize
        minsize = options.circle_minsize
        lsrc = numpy.zeros(len(SRCS), float)
        msrc = numpy.zeros(len(SRCS), float)
        nsrc = numpy.zeros(len(SRCS), float)
        beam_lm_src = numpy.zeros((2, len(SRCS)), float)
        for i, name in enumerate(SRCS):
            src = model.findSource(name)
            lsrc[i], msrc[i], nsrc[i] = Coordinates.radec_to_lmn(src.pos.ra, src.pos.dec, *radec0)
            if hasattr(src, "beam_lm"):
                beam_lm_src[:, i] = src.beam_lm
            else:
                beam_lm_src[:, i] = lsrc[i], msrc[i]

    # figure out optimal plot sizes
    if options.width or options.height:
        sz_as = sz_sa = (options.width, options.height)
    # figure out automatically depending on whether we have more sources or antennas
    else:
        # more sources than antennas
        if len(SRCS) < (len(ANTS) + 1):
            sz_as = (210, min(290, 30 * len(ANTS)))
            sz_sa = (290, min(210, 30 * len(SRCS)))
        else:
            sz_as = (290, min(210, 30 * len(ANTS)))
            sz_sa = (210, min(290, 30 * len(SRCS)))

    # initialize plot objects
    figplot = MultigridPlot(options)
    make_figure = figplot.make_figure

    skyplot = SkyPlot(options)
    make_skymap = skyplot.make_figure

    #
    # ============================ VARIOUS PER-CORRELATION PLOTS
    #
    if options.per_corr:
        for corr in XX, YY:
            if options.phase:
                if options.per_band:
                    make_figure(enumerate(ANTS), enumerate(SRCS),
                                lambda iant, isrc: [de_phase[spw, isrc, iant, corr, :] for spw in SPWS],
                                hline=0, ylock="row", mode=PLOT_MULTI,
                                suptitle="dE%s phase (deg) per band" % CORRS[corr],
                                save="dEphase_%s_perband" % CORRS[corr])
                # make_figure(enumerate(ANTS),enumerate(SRCS),
                # lambda iant,isrc:(numpy.mean(de_phase[:,isrc,iant,corr,:],0),
                # numpy.std(de_phase[:,isrc,iant,corr,:],0)),
                # hline=0,ylock="row",mode=PLOT_ERRORBARS,
                # suptitle="dE%s phase (deg) mean & stddev across all bands"%CORRS[corr],
                # save="dEphase_%s_mean"%CORRS[corr])
                make_figure(enumerate(SRCS), enumerate(ANTS),
                            lambda isrc, iant: (numpy.mean(de_phase[:, isrc, iant, corr, :], 0),
                                                numpy.std(de_phase[:, isrc, iant, corr, :], 0)),
                            hline=0, ylock="row", figsize=sz_sa, mode=PLOT_ERRORBARS,
                            suptitle="dE%s phase (deg) mean & stddev across all bands" % CORRS[corr],
                            save="dEphase_%s_mean" % CORRS[corr])
            if options.ampl:
                make_figure(enumerate(SRCS), enumerate(ANTS),
                            lambda isrc, iant: (numpy.mean(de_ampl[:, isrc, iant, corr, :], 0),
                                                numpy.std(de_ampl[:, isrc, iant, corr, :], 0)),
                            hline=1, ylock="row", figsize=sz_sa, mode=PLOT_ERRORBARS,
                            suptitle="dE%s amplitude mean & stddev across all bands" % CORRS[corr],
                            save="dEampl_%s_mean" % CORRS[corr])
            # make_figure(enumerate(SRCS),enumerate(ANTS),
            # lambda isrc,iant:(numpy.mean(dtec[:,isrc,iant,corr,:],0),
            # numpy.std(dtec[:,isrc,iant,corr,:],0)),
            # hline=0,ylock="row",figsize=sz_sa,mode=PLOT_ERRORBARS,
            # suptitle="dETEC(%s) mean & stddev across all bands"%CORRS[corr],
            # save="dETEC_%s_mean"%CORRS[corr])
            # make_figure(enumerate(ANTS[:-1]),enumerate(SRCS),
            # lambda iant,isrc:(dep0_res[:,isrc,iant,corr,:].mean(0),
            # dep0_res[:,isrc,iant,corr,:].std(0)),
            # hline=0,ylock="row",mode=PLOT_ERRORBARS,
            # suptitle="dE%s residual phase (w.r.t. slope fit) mean & stddev across all bands"%CORRS[corr],
            # save="dEphase_array_%s_res"%CORRS[corr])
            if options.ampl_slope:
                # make_figure(enumerate(ANTS[:-1]),enumerate(SRCS),
                # lambda iant,isrc:(dea0_res[:,isrc,iant,corr,:].mean(0),
                # dea0_res[:,isrc,iant,corr,:].std(0)),
                # hline=0,ylock="row",mode=PLOT_ERRORBARS,
                # suptitle="dE%s residual ampl (w.r.t. slope fit) mean & stddev across all bands"%CORRS[corr],
                # save="dEampl_array_%s_res"%CORRS[corr])
                make_figure(enumerate(SRCS), enumerate(ANTS[:-1]),
                            lambda isrc, iant: (dea0_res[:, isrc, iant, corr, :].mean(0),
                                                dea0_res[:, isrc, iant, corr, :].std(0)),
                            hline=0, ylock="row", mode=PLOT_ERRORBARS, figsize=sz_sa,
                            suptitle="dE%s residual ampl (w.r.t. slope fit) mean & stddev across all bands" % CORRS[
                                corr],
                            save="dEampl_array_%s_res" % CORRS[corr])

    #
    # ============================ DLM OFFSETS
    #
    dlm_offsets = dphase_lmfit = None;  # will be set if we fit them below
    if options.phase_slope_dlm:
        # compute best-fitting per-source offset
        # convert phase slope into phase offset on 0D baseline
        p0wl = dep0_b1 * (ANTX[-1] - ANTX[0]) * DEG
        p0wl /= 2 * math.pi * freq0[:, numpy.newaxis, numpy.newaxis] / 299792458.
        # take mean across SPWs (since we've eliminated dependence on freq),
        # p0wl is now NSRCxNTIME
        p0wl = p0wl.mean(0)
        # uv needs to be NTIMEx2
        uv = uvw[:, :2]
        # dlm will be NSRCx2
        dlm = dlm_offsets = numpy.zeros((len(SRCS), 2), float)
        print("=== Fitted dl,dm offsets:")
        for isrc, src in enumerate(SRCS):
            # flip sign of phase offset, since phase is -(ul+vm+w(n-1))
            x, res, rank, sing = linalg.lstsq(uv, -p0wl[isrc, :])
            print((
                "%8s=[%16.8g,%16.8g],     # %.2f\", %.2f\"" % (src, x[0], x[1], x[0] / ARCMIN * 60, x[1] / ARCMIN * 60)))
            dlm[isrc, :] = x
        # shape is now 2xNSRC
        dlm = dlm.transpose()
        # now make fitted phase curves
        ulvmwn = -(uvw[:, :2, numpy.newaxis] * dlm[numpy.newaxis, :, :]).sum(1)
        dphase_lmfit = (ulvmwn * 2 * math.pi * freq0.mean() / 299792458.) / (DEG * (ANTX[-1] - ANTX[0]))

        # make figure
        markers = []
        labels = []
        # set dlmscale so that longet arrow is .2 of max radius
        maxlm = math.sqrt((lsrc ** 2 + msrc ** 2).max())
        maxdlm = math.sqrt((dlm_offsets ** 2).sum(0).max())
        dlmscale = options.dlm_scale or 0.2 * maxlm / maxdlm
        # make l/m and dl/dm array (in arcmin)
        l0m0 = numpy.array([lsrc, msrc]).copy() / ARCMIN
        dldm = dlm_offsets.transpose().copy() * dlmscale / ARCMIN
        radius = options.radius;
        max(math.sqrt((l0m0 ** 2).sum(0).max()), math.sqrt(((l0m0 + dldm) ** 2).sum(0).max())) * 1.01
        # loop over sources
        for isrc, name in enumerate(SRCS):
            l0, m0 = l0m0[:, isrc]
            r = (math.sqrt((dlm_offsets[isrc, :] ** 2).sum()) / DEG) * 3600
            mark = ('arrow', list(l0m0[:, isrc]) + list(dldm[:, isrc]), dict(
                width=1, head_width=2, head_length=2,
            ))
            markers.append(mark)
            markers.append(('text', (l0, m0, "%.2f\"" % r), dict(
                fontsize=options.circle_label_fontsize,
                horizontalalignment='left' if dldm[0, isrc] > 0 else 'right',
                verticalalignment='top' if dldm[1, isrc] > 0 else 'bottom')))

        make_skymap(None, None, markers, labels=labels, radius=radius,
                    figsize=(210, 210), suptitle="Fitted l/m offsets", save="dE_lm_offsets")

    #
    # ============================ PHASE SLOPE PLOTS
    #
    # dep0_b0/1 will be NSPWxNSRCxNTIME
    if options.phase_slope:
        # compute rotational offset curves, if needed
        dr = options.phase_slope_rot_offset
        dt = options.phase_slope_time_offset
        if dr or dt:
            # dr: rotate lm by fixed angle
            if dr:
                # compute l1,m1,n1: rotated source lmn
                dr *= DEG
                cr, sr = math.cos(dr), math.sin(dr)
                l1 = lsrc * cr - msrc * sr
                m1 = lsrc * sr + msrc * cr
                n1 = numpy.sqrt(1 - l1 ** 2 - m1 ** 2)
                # compute dlmn: delta-lmn, shape is 3xNSRC
                dlmn = numpy.array([lsrc - l1, msrc - m1, nsrc - n1])
                # work out the corresponding phase difference (ul+vm+wn term)
                # shape is NTIMEx3xNSRC, then sum second axis, giving NTIMExNSRC
                ulvmwn = (uvw[:, :, numpy.newaxis] * dlmn[numpy.newaxis, :, :]).sum(1)
            # dt: introduce a time lag into the UVWs (in fractions of a timeslot)
            if dt:
                duvw = (uvw0[31::60] - uvw) * dt
                lmn = numpy.array([lsrc, msrc, nsrc - 1])
                # work out the corresponding phase difference (ul+vm+wn term)
                # shape is NTIMEx3xNSRC, then sum second axis, giving NTIMExNSRC
                ulvmwn = (duvw[:, :, numpy.newaxis] * lmn[numpy.newaxis, :, :]).sum(1)
            # now multiply (ul+vm+wn) by 2*pi/c
            dphase = ulvmwn * 2 * math.pi / 299792458.
            # now, convert this into a phase offset per NTIMExNSRC
            dphase = dphase[:, :] * freq0.mean()
            # and to deg/m
            dphase = dphase / (DEG * (ANTX[-1] - ANTX[0]))
            # if we've fitted a dphase corresponding to dlm, add it here
            if dphase_lmfit is None:
                plotfunc = lambda i, isrc: (
                    (None, numpy.mean(dep0_b1[:, isrc, :] * 1000, 0), numpy.std(dep0_b1[:, isrc, :] * 1000, 0)),
                    dphase[:, isrc] * 1000
                )
            else:
                plotfunc = lambda i, isrc: (
                    (None, numpy.mean(dep0_b1[:, isrc, :] * 1000, 0), numpy.std(dep0_b1[:, isrc, :] * 1000, 0)),
                    dphase_lmfit[:, isrc] * 1000, dphase[:, isrc] * 1000
                )
            # make the plot
            make_figure(enumerate(("",)), enumerate(SRCS), plotfunc,
                        ylock=False, mode=PLOT_MULTI,
                        suptitle="Fitted differential phase slope over array",
                        save="dEphase_array_slopes")
        else:
            if dphase_lmfit is None:
                plotfunc = lambda i, isrc: (
                    (None, numpy.mean(dep0_b1[:, isrc, :] * 1000, 0), numpy.std(dep0_b1[:, isrc, :], 0) * 1000),)
            else:
                plotfunc = lambda i, isrc: (
                    (None, numpy.mean(dep0_b1[:, isrc, :] * 1000, 0), numpy.std(dep0_b1[:, isrc, :] * 1000, 0)),
                    dphase_lmfit[:, isrc] * 1000
                )
            make_figure(enumerate(("",)), enumerate(SRCS), plotfunc,
                        ylock=False, mode=PLOT_MULTI,
                        suptitle="Fitted differential phase slope over array",
                        save="dEphase_array_slopes")

    #
    # ============================ RESIDUALS W.R.T. PHASE SLOPE
    #
    if options.phase_slope_res:
        make_figure(enumerate(SRCS), enumerate(ANTS[:-1]),
                    lambda isrc, iant: (dep0_res[:, isrc, iant, :].mean(0),
                                        dep0_res[:, isrc, iant, :].std(0)),
                    hline=0, ylock="row", mode=PLOT_ERRORBARS, figsize=sz_sa,
                    suptitle="dE residual phase (w.r.t. slope fit) mean & stddev across all bands",
                    save="dEphase_array_slopes_res")

    #
    # ============================ AMPLITUDE SLOPE PLOTS
    #
    if options.ampl_slope:
        dea0s = [dea0_b1[:, :, XX, :], dea0_b1[:, :, YY, :], dea0_b0[:, :, XX, :], dea0_b0[:, :, YY, :]]
        make_figure(enumerate(("X slope 1/m", "Y slope 1/m", "X offset", "Y offset")), enumerate(SRCS),
                    lambda i, isrc: (numpy.mean(dea0s[i][:, isrc, :], 0),
                                     numpy.std(dea0s[i][:, isrc, :], 0)),
                    ylock=False, mode=PLOT_ERRORBARS,
                    suptitle="Fitted differential ampl slope over array",
                    save="dEampl_array_slopes")

    #
    # ============================ POLARIZATION PLOT
    #
    if options.pol:
        make_figure(enumerate(SRCS), enumerate(ANTS),
                    lambda isrc, iant: (
                    numpy.mean(de_fpol[:, isrc, iant, :], 0), numpy.std(de_fpol[:, isrc, iant, :], 0)),
                    hline=0, ylock="row", mode=PLOT_ERRORBARS,
                    suptitle="Fractional instrumental polarization, average across all bands",
                    save="dEpol_mean")
        if options.per_band:
            make_figure(enumerate(SRCS), enumerate(ANTS),
                        lambda isrc, iant: [de_fpol[spw, isrc, iant, :] for spw in SPWS],
                        hline=0, ylock="row", mode=PLOT_MULTI,
                        suptitle="Fractional instrumental polarization per band",
                        save="dEpol_mean_sa_perband")

    #
    # ============================ dE PHASE PLOTS
    #
    if options.phase:
        de_mean_phase = de_phase.mean(3)
        make_figure(enumerate(SRCS), enumerate(ANTS[:-1]),  # omit MEAN antenna since it's 1 by definition
                    lambda isrc, iant: (numpy.mean(de_mean_phase[:, isrc, iant, :], 0),
                                        numpy.std(de_mean_phase[:, isrc, iant, :], 0)),
                    hline=0, ylock="row", figsize=sz_sa, mode=PLOT_ERRORBARS,
                    suptitle="dE phase (deg) mean & stddev across all bands",
                    save="dEphase_mean")

    #
    # ============================ dE AMPLITUDE PLOTS
    #
    if options.ampl:
        if options.pe:
            de_pe = numpy.zeros(de_unnorm.shape, float)
            de_peshape = numpy.zeros(de_unnorm.shape, float)
            # pe_dlm is 2 x NSPW x NANT x NTIME
            # beam_lm_src is 2 x NSRC
            # compute theoretical gain [NSRC]
            r0 = numpy.sqrt((beam_lm_src ** 2).sum(0))
            # gain0 is NSRC x NSPW
            gain0 = r0[:, numpy.newaxis] * freq0[numpy.newaxis, :] * 6.5e-8
            gain0 = numpy.maximum(numpy.cos(gain0) ** 3, 0.1)
            # lm is then the actual source offset: this is 2 x NSRC x NSPW x NANT x NTIME
            #      lm = numpy.zeros((2,len(SRCS),len(ANTS),NTIMES),float)
            lm = beam_lm_src[:, :, numpy.newaxis, numpy.newaxis, numpy.newaxis] - \
                 pe_dlm[:, numpy.newaxis, :, :, :]
            # convert to r=sqrt(l^2+m^2)*freq*C: NSRC x NSPW x NANT x NTIME
            pe_r = numpy.sqrt((lm ** 2).sum(0));  ## sqrt(l^2+m^2)
            # multiply by C and freq
            pe_r *= freq0[numpy.newaxis, :, numpy.newaxis, numpy.newaxis] * 6.5e-8
            # convert to gain
            beamgain_pe = numpy.maximum(numpy.cos(pe_r) ** 3, .1)
            beamgain_pe = beamgain_pe / gain0[:, :, numpy.newaxis, numpy.newaxis]
            # if we have shape, compute also beamgain due to shape
            if pe_beam_sizes > 0:
                beamgain_pe_shape = numpy.maximum(numpy.cos(pe_r / pe_bsz[0, 0, numpy.newaxis, :, :, :]) ** 3, .1)
                beamgain_pe_shape = beamgain_pe_shape / gain0[:, :, numpy.newaxis, numpy.newaxis]
                # make plotting func
                plotfunc = lambda isrc, iant: (
                    (None, numpy.mean(de_unnorm[:, isrc, iant, :], 0), numpy.std(de_unnorm[:, isrc, iant, :], 0)),
                    beamgain_pe[isrc, :, iant, :].mean(0),
                    beamgain_pe_shape[isrc, :, iant, :].mean(0),
                )
            else:
                plotfunc = lambda isrc, iant: (
                    (None, numpy.mean(de_unnorm[:, isrc, iant, :], 0), numpy.std(de_unnorm[:, isrc, iant, :], 0)),
                    beamgain_pe[isrc, :, iant, :].mean(0),
                )
            make_figure(enumerate(SRCS), enumerate(ANTS[:-1]),
                        plotfunc,
                        hline=1, ylock="row", figsize=sz_sa, mode=PLOT_MULTI,
                        suptitle="||dE|| mean & stddev across all bands, plus fitted ||dE|| due to pointing error",
                        save="dE_mean")
        else:
            make_figure(enumerate(SRCS), enumerate(ANTS[:-1]),  # omit MEAN antenna since it's 1 by definition
                        lambda isrc, iant: (
                        numpy.mean(de_renorm[:, isrc, iant, :], 0), numpy.std(de_renorm[:, isrc, iant, :], 0)),
                        hline=1, ylock="row", figsize=sz_sa, mode=PLOT_ERRORBARS,
                        suptitle="Renormalized ||dE|| mean & stddev across all bands",
                        save="dE_mean")
        if options.per_band:
            make_figure(enumerate(SRCS), enumerate(ANTS),
                        lambda isrc, iant: [de_renorm[spw, isrc, iant, :] for spw in SPWS],
                        hline=1, ylock="row", figsize=sz_sa, mode=PLOT_MULTI,
                        suptitle="Renormalized ||dE|| per band",
                        save="dE_mean_perband")

    # make_figure(enumerate(SRCS),enumerate(ANTS),
    # lambda isrc,iant:(numpy.mean(de_norm[:,isrc,iant,:],0),numpy.std(de_norm[:,isrc,iant,:],0)),
    # hline=1,ylock="row",figsize=sz_sa,mode=PLOT_ERRORBARS,
    # suptitle="||dE|| mean & stddev across all bands",
    # save="dE_mean")

    # make_figure(enumerate(ANTS),enumerate(SRCS),
    # lambda iant,isrc:[numpy.mean(abs(de[:,isrc,iant,corr,:]),0) for corr in XX,YY],
    # hline=1,ylock="row",mode=PLOT_MULTI,
    # suptitle="|dExx|,|dEyy| averaged across all bands",
    # save="dE_mean_xx_yy")

    #
    # ============================ ROGUES GALLERY AMPLITUDES
    #
    output_prefix = options.output_prefix + "_" if options.output_prefix else ""
    if options.circle_ampl:
        # de_renorm is NSPWxNSRCxNANTxNTIME
        # reshape as NSRCxNANTxNSPWxNTIME
        de_r1 = de_renorm.copy().transpose((1, 2, 0, 3))
        demean = de_r1.mean(2).mean(2)
        # stddev is sqrt(s1^2+s2^2), where s1 is mean error, and s2 is error of mean
        destd = de_r1.std(2).mean(2) / math.sqrt(NTIMES - 1)
        #    destd  = numpy.sqrt((de_r1.std(2).mean(2)/math.sqrt(NTIMES-1))**2 + de_r1.mean(2).std(2)**2)
        delog = numpy.where(demean >= 1, numpy.sqrt(demean - 1), -numpy.sqrt(1 - demean))

        antplots = []

        for iant, ant in list(enumerate(ANTS))[:-1]:
            # make figure
            markers = []
            labels = []
            for name, log, mean, std in zip(SRCS, delog[:, iant], demean[:, iant], destd[:, iant]):
                color = "blue" if log >= 0 else "red"
                mark = dict(marker='o', markersize=minsize + (maxsize - minsize) * (abs(log) / maxabsdelog),
                            markeredgecolor=color, markerfacecolor='None')
                sigmas = abs(1 - mean) / std
                if sigmas >= 3:
                    mark['markerfacecolor'] = color
                else:
                    sigmas = max(sigmas, 1)
                    mark['markeredgewidth'] = options.circle_linewidth * ((sigmas - 1) / 2.) + 1
                #        elif sigmas >= 1:
                #          mark['markeredgewidth'] = 2*options.circle_linewidth
                #        else:
                #          mark['markeredgewidth'] = 1
                markers.append(mark)
                if options.names_only:
                    labels.append(name)
                else:
                    labels.append("%s %.2f+-%.2f" % (name, mean, std))

            antplots.append(make_skymap(lsrc, msrc, markers, labels=labels,
                                        figsize=(210, 210), suptitle="Average ||dE||, RT%s" % ant,
                                        save="dE_ant%s" % ant))

        # try to make summary plot
        if options.output_type.upper() == "PNG":
            try:
                import PIL
                import PIL.Image
            except ImportError:
                print("Python Imaging Library (PIL) not found, can't make the antenna rogues' gallery")
                PIL = None
            if PIL:
                ncol = 5;  # number of plots per column
                plotsize = 256;  # column width, in pixels
                ncol = min(ncol, len(antplots))
                nrow, rem = divmod(len(antplots), ncol)
                if rem:
                    nrow += 1
                # make image
                img = PIL.Image.new('RGBA', (ncol * plotsize, nrow * plotsize))
                for i, plot in enumerate(antplots):
                    subimg = PIL.Image.open(plot)
                    subimg = subimg.resize((plotsize, plotsize), PIL.Image.ANTIALIAS)
                    y0, x0 = divmod(i, ncol)
                    img.paste(subimg, (x0 * plotsize, y0 * plotsize))
                outname = "%sdE_ant_gallery.png" % output_prefix
                img.save(outname, "PNG")
                print(("Wrote", outname))

    #
    # ============================ ROGUES GALLERY ANIMATION
    #
    if circle_ampl_ant is not None:
        # make animation for given antenna
        antname = ANTS[circle_ampl_ant]

        # de_renorm is NSPWxNSRCxNANTxNTIME
        demean = de_renorm[:, :, circle_ampl_ant, :].mean(0)
        destd = de_renorm[:, :, circle_ampl_ant, :].std(0)
        delog = numpy.sqrt(demean);  # numpy.log10(demean)
        delog = numpy.where(demean >= 1, numpy.sqrt(demean - 1), -numpy.sqrt(1 - demean))

        frames = []
        for k in range(NTIMES):
            markers = []
            labels = []
            for name, log, mean, std in zip(SRCS, delog[:, k], demean[:, k], destd[:, k]):
                color = "#2b60d3" if log >= 0 else "red"
                mark = dict(marker='o', markersize=minsize + (maxsize - minsize) * (abs(log) / maxabsdelog),
                            markeredgecolor=color, markerfacecolor=color)
                markers.append(mark)
                if options.names_only:
                    labels.append(name)
                else:
                    labels.append("%s %.2f+-%.2f" % (name, mean, std))

            frames.append(make_skymap(lsrc, msrc, markers, labels=labels,
                                      figsize=(210, 210),
                                      suptitle="Average ||dE||, antenna %s, time slice %d" % (antname, k),
                                      save="%sdE_ant%s_%03d" % (output_prefix, antname, k)))

        # try to make animation
        if options.output_type.upper() == "PNG":
            for f in frames:
                os.system("pngtopnm %s | ppmquant 256 | ppmtogif >%s.gif" % (f, f))
            outname = "%sdE_ant%s_anim.gif" % (output_prefix, antname)
            os.system("gifsicle -d 15 -l0 --colors 256 %s >%s" %
                      (" ".join([f + ".gif" for f in frames + frames[-1::-1]]),
                       outname))
            os.system("rm -f %s" % " ".join([f + ".gif" for f in frames]))
            print(("Wrote", outname))

    #
    # ============================ UPDATE LSM
    #
    if options.update_lsm:
        if lsm_timestamp >= mep_timestamp:
            if input("Your LSM file appears to be newer than the dE solutions. Really update (y/n)? "
                     ).lower()[:1] != "y":
                print("LSM update cancelled.")
                sys.exit(1)
        # update sources
        print("=== Updating model sources")
        A = (dex ** 2 + dey ** 2) / 2
        B = (dex ** 2 - dey ** 2) / 2
        for isrc, name in enumerate(SRCS):
            for src in model.sources:
                if src.name.startswith(name):
                    if hasattr(src.flux, 'Q'):
                        I, Q = src.flux.I, src.flux.Q
                        src.flux.I = A[isrc] * I + B[isrc] * Q
                        src.flux.Q = A[isrc] * Q + B[isrc] * I
                        print(("%8s I=%f Q=%f --> I=%f Q=%f" % (src.name, I, Q, src.flux.I, src.flux.Q)))
                    if dlm_offsets is not None:
                        ra, dec = src.pos.ra, src.pos.dec
                        l, m, n = Coordinates.radec_to_lmn(ra, dec, *radec0)
                        l += dlm_offsets[isrc, 0]
                        m += dlm_offsets[isrc, 1]
                        src.pos.ra, src.pos.dec = Coordinates.lm_to_radec(l, m, *radec0)
                        print(("%8s position %.8f,%.8f --> %.8f %.8f" % (src.name, ra, dec, src.pos.ra, src.pos.dec)))
        newname = "updated-" + lsm_filename
        model.save(newname)
        print(("Wrote updated sky model", newname))

    if options.output_type.upper() == "X11":
        from pylab import plt

        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path
import math
import sys
import re
import warnings

import numpy
import numpy.ma
from past.builtins import cmp

import Owlcat

from Owlcat import Parsing

COMPLEX_CIRCLE = "cc";  # identifier for the complex circle plot type
COMPLEX_CIRCLE_MEAN = "ccm";  # identifier for the complex circle plot type


#
# NB: these need to be constructed on-the-fly based on polarization info from the MS.
# Otherwise we don't support circular without kludging
#

def cca_datafunc(data):
    while data.ndim > 1:
        data = data.mean(0)
    return data


PT_DATATRACK = 0
PT_FLAGTRACK = 1
PT_CC = 2

# Each plotter func is a tuple of
#   label,description,callable,plot_type
# The 'callable' transforms data into plottables
# If type is PT_DATATRACK, callable(visibilities) should return an array of real plottables
# the same shape as the visibilities masked_array.
# If type is PT_FLAGTRACK, callable(flags,axis) should return an array of flag densities
# along the specified axis. The returned array should be reduced along 'axis'.
# If type is PT_CC, callable(visibilities) should return an array of complex points to plot.
# The last axis is correlation.
Plotters = [
    ("I", "Stokes I", lambda data: (abs(data[..., 0]) + abs(data[..., 3])) / 2, PT_DATATRACK),
    ("Q", "Stokes Q", lambda data: (abs(data[..., 0]) - abs(data[..., 3])) / 2, PT_DATATRACK),
    ("U", "Stokes U", lambda data: (data[..., 1].real + data[..., 2].real) / 2, PT_DATATRACK),
    ("V", "Stokes V", lambda data: (data[..., 1].imag - data[..., 2].imag) / 2, PT_DATATRACK),
    ("cc", "Complex circle plot", lambda data: data, PT_CC),
    ("cca", "Complex circle averages plot", cca_datafunc, PT_CC),
    ("flags_I", "I/Q flag density",
     lambda flags, meanaxis:
     (flags[..., 0] | flags[..., 3]).mean(meanaxis), PT_FLAGTRACK),
    ("flags_all", "2x2 flag density",
     lambda flags, meanaxis:
     (flags[..., 0] | flags[..., 1] | flags[..., 2] | flags[..., 3]).mean(meanaxis), PT_FLAGTRACK),
]

for icorr, corr in enumerate(("XX", "XY", "YX", "YY")):
    Plotters += [
        (corr, corr + " amplitude", lambda data, i=icorr: abs(data[..., i]), PT_DATATRACK),
        (corr + "phase", corr + " phase", lambda data, i=icorr: numpy.ma.masked_array(numpy.angle(data[..., i]),
                                                                                      data[..., i].mask), PT_DATATRACK),
        (corr + "r", corr + " real", lambda data, i=icorr: data[..., i].real, PT_DATATRACK),
        (corr + "i", corr + " imag", lambda data, i=icorr: data[..., i].imag, PT_DATATRACK),
        ("flags_" + corr, corr + " flag density",
         lambda flags, meanaxis, i=icorr:
         flags[..., i].mean(meanaxis), PT_FLAGTRACK),
    ]

if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser, OptionGroup

    parser = OptionParser(usage="""%prog: [options] MS column:plot [column:plot ...]""",
                          description="""Makes plots of a column in the MS. Plots are specified as "column:plottype". If
column is not given, it defaults to the previous column, or CORRECTED_DATA. Use --list-plots to get
more information on available plots. If no plots are specified, CORRECTED_DATA:I is plotted by default.
""")

    parser.add_option("--list-plots", action="store_true",
                      help="list available plot types and exit")
    parser.add_option("-L", "--channels", dest="freqslice", type="string",
                      help="channel selection: single number or start:end[:step] to select channels start through end-1, "
                           "or start~end[:step] to select channels start through end, with an optional stepping.")
    parser.add_option("-T", "--timeslots", dest="timeslice", type="string",
                      help="timeslot selection by number, same format as channels.")
    parser.add_option("-I", "--ifrs", dest="ifrs", type="string",
                      help="subset of interferometers to plot. Use \"-I help\" to get help on selecting ifrs.")
    parser.add_option("-D", "--ddid", dest="ddid", type="string",
                      help="DATA_DESC_ID to plot. Default is first DDID found in MS. Use comma-separated list "
                           "(or 'all') to plot multiple DDIDs.")
    parser.add_option("-F", "--field", dest="field", type="int",
                      help="FIELD_ID to plot. Default is all (this may produce some strange plots for multiple-field MSs.)")
    parser.add_option("-Q", "--taql", dest="taql", type="string",
                      help="additional TaQL selection to restrict data subset.")
    parser.add_option("-f", "--flagmask", metavar="FLAGS", dest="flagmask", type="string",
                      help="flagmask to apply to the BITFLAG column to get flagged data points"
                           "default is ALL, use 0 to ignore flags.")

    group = OptionGroup(parser, "Arranging your plots")
    group.add_option("--x-time", dest="xaxis", action="store_const", const=0,
                     help="use time for X axis and average over channels (default)")
    group.add_option("--x-freq", dest="xaxis", action="store_const", const=1,
                     help="use frequency for X axis and average over timeslots")
    group.add_option("--x-grid", type='int', action="append",
                     help="sets X grid interval. Use 0 to disable grid. Use twice to set major and minor intervals.")
    group.add_option("--no-y-grid", action="store_true",
                     help="disable Y axis grid")

    PLOT, DDID, IFR = group_choices = ["plot", "ddid", "ifr"]
    group.add_option("-P", "--page", metavar="|".join(group_choices), action="append",
                     type="choice", choices=group_choices,
                     help="put different plots, DDIDs and/or interferometers on "
                          "separate pages. Can be given multiple times, this will determine "
                          "page order. Default is to put DDIDs and plots on different pages, and to "
                          "stack interferometers.")
    group.add_option("-S", "--stack", metavar="|".join(group_choices), action="append",
                     type="choice", choices=group_choices,
                     help="stack different plots, DDIDs and/or interferometers on "
                          "a single page. Can be given multiple times, this will determine "
                          "stacking order.")
    choices = group_choices[1:]
    group.add_option("-A", "--average", metavar="|".join(choices), action="append",
                     type="choice", choices=choices,
                     help="average DDIDs and/or interferometers together. Use twice "
                          "to average both.")
    group.add_option("-G", "--group-redundant", action="store_true",
                     help="group each group of redundant baselines into a single plot.")
    group.add_option("--ppp", dest="ppp", metavar="N", type="int",
                     help="maximum number of plots to stack per page.")
    group.add_option("--offset-std", dest="offset_std", metavar="X", type="float",
                     help="vertical offset between stacked plots, in units of the median stddev. "
                          "Default is 10. NB: for flag-density plots, unit is 0.11*maxdensity, so the "
                          "standard offset of 10 produces plots spaced at 110% of the max value.")
    parser.add_option("--offset", type="float",
                      help="vertical offset between stacked plots, in absolute units. This overrides --offset-std.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Plot labels")
    group.add_option("--no-label-plot", dest="label_plot", action="store_false",
                     help="do not include plot type in labels (when stacking)")
    group.add_option("--no-label-ddid", dest="label_ddid", action="store_false",
                     help="do not include DDID in labels (when stacking)")
    group.add_option("--no-label-ifr", dest="label_ifr", action="store_false",
                     help="do not include IFR in labels (when stacking)")
    group.add_option("--no-label-baseline", dest="label_baseline", action="store_false",
                     help="do not include baseline length in labels")
    group.add_option("--no-label-mean", dest="label_mean", action="store_false",
                     help="do not include mean value in labels")
    group.add_option("--no-label-stddev", dest="label_stddev", action="store_false",
                     help="do not include standard deviation in labels")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Output options")
    group.add_option("--title", dest="title", type="string", action="append",
                     help="provide a custom plot title. When you have multiple pages, you may "
                          "give this option serveral times.")
    group.add_option("-o", "--output", metavar="FILE", dest="output", type="string",
                     help="plot output to FILE. If not specified, plot will be "
                          "shown in a window. File format is determined by extension, see matplotlib documentation "
                          "for supported formats. At least .png, .pdf, .ps, .eps and .svg are supported.")
    group.add_option("--dpi", dest="resolution", type="int", metavar="DPI",
                     help="plot resolution for output to FILE (default is %default)")
    group.add_option("--size", metavar="WxH", dest="figsize", type="string",
                     help="set figure size, in cm. Default is '%default'.")
    group.add_option("--papertype", dest="papertype", type="string",
                     help="set paper type (for .ps output only.) Prefix with '/' for "
                          "landscape mode. Default is '%default', but can also use e.g. 'letter', 'a3', etc.")
    parser.add_option_group(group)

    parser.set_defaults(output="", xaxis=0, ddid='first', field=None,
                        resolution=300, ppp=0, papertype='a4', figsize="21x29", offset_std=10, offset=None,
                        x_grid=[],
                        flag_mask=None,
                        label_plot=True,
                        label_ddid=True,
                        label_ifr=True,
                        label_baseline=True,
                        label_mean=True,
                        label_stddev=True,
                        page=[], stack=[], average=[])

    (options, args) = parser.parse_args()

    if not options.ppp:
        options.ppp = 120 if not options.group_redundant else 20

    # print help on plotters
    if options.list_plots:
        print("Available plot types:\n")
        for p in Plotters:
            print("   %-16s%s" % (p[0], p[1]))
        print("""
By default, specifying "DATA:XX" is the same as "DATA:XX.mean", producing a plot of mean |XX|
in time or frequency (depending on choice of X axis). Use "DATA.mean:XX" to plot |mean XX|
instead (i.e. mean visibilities, not mean amplitudes!) Use "DATA:XX.stddev" or "DATA.stddev:XX"
to plot standard deviations. Other options are ".sum", ".max" and ".min".

Note that flag-density plots do not require a data column to be specified, since the
BITFLAG/FLAG columns are shared among all data columns.
""")
        sys.exit(0)

    # turn list of plotters into dict
    Plotters = dict([(p[0], p[1:]) for p in Plotters])

    # figure out plot arrangements, insert defaults for things that are not specified explicitly
    allopt = options.page + options.stack + options.average
    if PLOT not in allopt:
        options.page.insert(0, PLOT)
    if DDID not in allopt:
        options.page.insert(0, DDID)
    if IFR not in allopt:
        options.stack.insert(0, IFR)
    # check that there's no double entries
    for what in [PLOT, DDID, IFR]:
        if len([opt for opt in allopt if opt == what]) > 1:
            parser.error("Conflicting --page/--stack/--average options for '%s'.")

    # get figsize
    try:
        w, h = list(map(int, options.figsize.split('x', 1)))
        figsize = (w * 10, h * 10)
    except:
        parser.error("Invalid --size setting '%s'" % options.figsize)

    # get MS name
    if not args:
        parser.error("MS not specified. Use '-h' for help.")
    msname = args[0]
    print("===> Attaching to MS %s" % msname)
    ms = Owlcat.table(msname)

    # get flagmask, use a Flagger for this
    import Owlcat.Flagger
    from Owlcat.Flagger import Flagger
    flagger = Flagger(msname)

    if options.flagmask == "0":
        flagmask = 0
        print("===> Flagmask 0, ignoring all flags")
    elif options.flagmask is not None:
        flagmask = flagger.lookup_flagmask(options.flagmask)
        print("===> Flagmask is %s (you specified '%s')" % (flagger.flagmaskstr(flagmask), options.flagmask))
        if not flagmask & flagger.LEGACY:
            print("===> NB: legacy FLAG/FLAG_ROW columns will be ignored with this flagmask")
    else:
        flagmask = flagger.BITMASK_ALL | flagger.LEGACY
        print("===> Using all flags")

    # parse slice specs
    try:
        freqslice = Parsing.parse_slice(options.freqslice)
        print("===> Channel selection is ", freqslice)
    except:
        parser.error("Invalid -L/--channels option. Start:end[:step] or start~end[:step] expected.")
    try:
        timeslice = Parsing.parse_slice(options.timeslice)
        print("===> Timeslot selection is ", timeslice)
    except:
        parser.error("Invalid -T/--timeslots option. Start:end[:step] start~end[:step] expected.")

    # make list of plots
    plots = []
    has_flagplots = False
    column0 = "CORRECTED_DATA"
    if column0 not in ms.colnames():
        column0 = "DATA"
    # go through list of arguments, or default list
    for arg in (args[1:] or ["I"]):
        # parse as "[column[.reduce]:]plot[.reduce]"
        m = re.match('^(\w+)(\.(\w+))?(:(\w+)(.(\w+))?)?$', arg)
        if not m:
            parser.error("'%s': invalid argument" % arg)
        g = m.groups()
        # groups 0-2 are the first field, 3-6 are the second field
        # if 3 didn't match, then we only have one field, which we treat as a 'plot' designator (I, XX, etc.)
        if g[3]:
            column, datareduce, plot, plotreduce = g[0], g[2], g[4], g[6]
        else:
            column, datareduce, plot, plotreduce = None, None, g[0], g[2]
        # the reduction function applies to visibilities or to plottable values
        if plotreduce and datareduce:
            parser.error("'%s': can't specify two .funcs" % arg)
        # check that function is valid
        for func in datareduce, plotreduce:
            if func not in [None, 'mean', 'std', 'min', 'max', 'sum', 'product']:
                parser.error("'%s': '%s' is not a valid reduction function" % (arg, func))
        # lookup plotter
        plotdesc, func, plot_type = Plotters.get(plot, (None, None, None))
        if func is None:
            parser.error("'%s': unknown plot type '%s'" % (arg, plot))
        # more sanity checks
        if plot_type is PT_FLAGTRACK:
            has_flagplots = True
            if plotreduce or datareduce:
                parser.error("'%s': can't use '.%s' with flagplots" % (arg, (plotreduce or datareduce)))
            if column:
                parser.error("'%s': can't specify a column for flagplots" % arg)
        else:
            column0 = column or column0
            if column0 not in ms.colnames():
                parser.error("MS does not contain a %s column" % column)
            # set default reduction, if nothing else is specified
            if not plotreduce and not datareduce:
                plotreduce = 'mean'
            plotdesc = " ".join(filter(bool, [datareduce, column0, plotreduce, plotdesc]))
        # add to list of plots
        plots.append((plot, plotdesc, func, plot_type, column0, datareduce, plotreduce))

    # collect applicable TaQL queries here
    taqls = []

    # get IFR set
    import Meow.IfrSet

    ifrset = Meow.IfrSet.from_ms(ms)
    tot_ifrs = len(ifrset.ifrs())
    if options.ifrs:
        if options.ifrs == "help":
            # print help string, but trim away RTF tags
            print(re.sub("<[^>]+>", "", ifrset.subset_doc).replace("&lt;", "<").replace("&gt;", ">"))
            sys.exit(0)
        ifrset = ifrset.subset(options.ifrs)
        taqls.append(ifrset.taql_string())
        print("===> Selected %d of %d interferometers " % (len(ifrset.ifrs()), tot_ifrs))
    else:
        print("===> Selected all %d interferometers " % tot_ifrs)

    # select DDIDs
    ddid_tab = Owlcat.table(ms.getkeyword('DATA_DESCRIPTION'))
    # default ('first') is to use first DDID in MS
    ddid_str = options.ddid.strip().lower()
    if ddid_str == 'first':
        ddids = [ms.getcol('DATA_DESC_ID', 0, 1)[0]]
        print("===> Using first DATA_DESC_ID (%d)" % ddids[0])
    # else see if it's a single int
    elif re.match('^\d+$', ddid_str):
        ddids = [int(options.ddid)]
    # else parse as list of ints, or 'all'. In this case we need extra info from the tables
    else:
        if ddid_str == "all":
            ddids = list(range(ddid_tab.nrows()))
        else:
            try:
                ddids = list(map(int, ddid_str.split(',')))
            except:
                parser.error("Invalid -D/--ddid option: %s" % ddid_str)

    # Get data shapes by reading subtables. I used to just use:
    #     datashape = [ ms.nrows() ] + list(ms.getcoldesc('DATA')['shape'])
    # but this is not robust, since some MSs will not have a fixed-shape DATA column. So, read the subtables.
    spwids = ddid_tab.getcol('SPECTRAL_WINDOW_ID')
    polids = ddid_tab.getcol('POLARIZATION_ID')
    corrs = Owlcat.table(ms.getkeyword('POLARIZATION')).getcol('CORR_TYPE')
    spw_tab = Owlcat.table(ms.getkeyword('SPECTRAL_WINDOW'))
    ref_freq = spw_tab.getcol('REF_FREQUENCY')
    nchan = spw_tab.getcol('NUM_CHAN')

    # select field
    if options.field is not None:
        taqls.append("FIELD_ID==%d" % options.field)
        print("===> Selecting FIELD_ID %d" % options.field)

    # select TaQL
    if options.taql:
        taqls.append(options.taql)
        print("===> Additional TaQL query is \"%s\"" % options.taql)

    # apply accumulated selection
    if taqls:
        ms = ms.query("( " + " ) && ( ".join(taqls) + " )")
        print("===> Selected %d rows from MS" % ms.nrows())
    if not ms.nrows():
        print("""MS selection is empty. You may have specified it incorrectly: please check your
DATA_DESC_ID (option -D/--ddid), field (-F/--field), interferometer subset (-I/--ifrs)
and/or TaQL query (-Q/--taql) options. Or was your MS empty to begin with?""")
        sys.exit(1)

    # get axis indices
    xaxis = options.xaxis
    meanaxis = 1 if xaxis == 0 else 0

    # figure out label format
    labels = []
    if options.label_plot and PLOT in options.stack:
        labels.append("%(plot)s")
    if options.label_ddid and DDID in options.stack:
        labels.append("d#%(ddid)d")
    if options.label_ifr and IFR in options.stack:
        labels.append("%(ifr)s")
    if options.label_baseline and IFR in options.stack:
        labels.append("%(baseline)dm")
    if options.label_mean:
        labels.append("mean=%(mean).3g")
    if options.label_stddev:
        labels.append("std=%(stddev).3g")
    label_format = " ".join(labels)

    # Make list of trackkeys
    # A trackkey is (nplot,ddid,ifr); ddid or ifr is None if averaging
    average_ddids = DDID in options.average
    average_ifrs = IFR in options.average
    # ranges for each trackkey
    keyranges = [list(range(len(plots))),
                 [None] if average_ddids else ddids,
                 [None] if average_ifrs else [(px[0], qx[0]) for px, qx in ifrset.ifr_index()]]

    # set non-interactive backend, if -o is in effect
    import matplotlib

    if options.output:
        matplotlib.use('agg')
    else:
        matplotlib.use('qt4agg')

    import Owlcat.Plotting

    # A PlotCollection holds one set of plot tracks.
    # This is a dict of trackkey:PC.
    # PC's are shared by multiple tracks  according to the page/stack/average settings.
    plotcolls = {}


    # Helper function to get a PC associated with the given trackkey
    # Creates one if it doesn't exist, and inserts it into the plotcolls dict
    # for every other track in the stacked set.
    def get_plot_collection(track, plottype):
        pc = plotcolls.get(track)
        if pc:
            return pc
        # make new PlotCollection
        if plottype is PT_CC:
            pc = plotcolls[track] = Owlcat.Plotting.ComplexCirclePlot(options)
        elif options.group_redundant:
            pc = plotcolls[track] = Owlcat.Plotting.PlotCollectionSep(options)
        else:
            pc = plotcolls[track] = Owlcat.Plotting.PlotCollection(options)
        #    print "New pc for",track
        # associate it with every track that will go on it, by looping over stacked keys
        for plot1 in (keyranges[0] if PLOT in options.stack else [track[0]]):
            for ddid1 in (keyranges[1] if DDID in options.stack else [track[1]]):
                for ifr1 in (keyranges[2] if IFR in options.stack else [track[2]]):
                    plotcolls[plot1, ddid1, ifr1] = pc
        #          print "also for ",(plot1,ddid1,ifr1)
        return pc


    # set of IFRs for which (non-flagged) data is actually found
    active_ifrs = set()
    labelattrs = {};  # dict of plot label attrs, updated in each loop

    # now loop over DDIDs
    for ddid in ddids:
        labelattrs['ddid'] = ddid

        subms = ms.query("DATA_DESC_ID==%d" % ddid)
        datashape = [subms.nrows(), nchan[spwids[ddid]], len(corrs[polids[ddid]])]

        print("===> Processing DATA_DESC_ID %d (%d MHz): %dx%d by %d rows" % (
            ddid, round(ref_freq[spwids[ddid]] * 1e-6), datashape[1], datashape[2], datashape[0]))

        if not subms.nrows():
            continue
        # get boolean flags
        if flagmask & flagger.LEGACY:
            flagcol = subms.getcol('FLAG')
            # merge in FLAG_ROW column
            flagrowcol = subms.getcol('FLAG_ROW')
            flagcol |= flagrowcol[:, numpy.newaxis, numpy.newaxis]
        else:
            flagcol = numpy.zeros(datashape, bool)
        # get bitflags
        bitflags = flagmask & flagger.BITMASK_ALL
        if bitflags:
            if 'BITFLAG' in ms.colnames():
                bf = subms.getcol('BITFLAG')
                flagcol |= ((bf & bitflags) != 0)
            if 'BITFLAG_ROW' in ms.colnames():
                bfr = subms.getcol('BITFLAG_ROW')
                flagcol |= ((bfr & bitflags) != 0)[:, numpy.newaxis, numpy.newaxis]
        flagcol = flagcol[:, freqslice, :]
        nf = flagcol.sum()
        print("===> %d of %d (%.2g%%) visibilities are flagged " % (nf, flagcol.size, (nf / float(flagcol.size)) * 100))

        # get antenna indices, and make dict per-ifr masks
        a1, a2 = subms.getcol('ANTENNA1'), subms.getcol('ANTENNA2')
        ifr_rows = dict(
            [((p, q), numpy.where((a1 == p) & (a2 == q))[0]) for (p, plab), (q, qlab) in ifrset.ifr_index()])
        # visibility columns will be read into here, as demanded by the individual plots
        datacols = {}

        colname0 = None
        for iplot, (plotwhat, plotdesc, plotfunc, plot_type, colname, datareduce, plotreduce) in enumerate(plots):
            if colname0 and colname and colname != colname0:
                labelattrs['plot'] = "%s:%s" % (colname, plotwhat)
            else:
                labelattrs['plot'] = plotwhat
            colname0 = colname
            print("===> Making plot", plotdesc)
            # read in data column, if not already read
            if colname:
                # do we already have a reduced version of the column cached?
                dc = datacols.get((colname, datareduce), None)
                if dc is None:
                    # no, do we have an unreduced version cached?
                    dc = datacols.get((colname, None), None)
                    if dc is None:
                        # ok, read & cache column
                        dc = subms.getcol(colname)[:, freqslice, :]
                        mask = flagcol | (~numpy.isfinite(dc))
                        dc = datacols[colname, None] = numpy.ma.masked_array(dc, mask, fill_value=complex(0, 0))
                    # reduce and cache reduced version
                    if datareduce:
                        dc = datacols[colname, datareduce] = getattr(dc, datareduce)(meanaxis)

            # split up into per-baseline arrays, and accumulate data tracks
            baselines = set()
            for (p, plab), (q, qlab) in ifrset.ifr_index():
                idx = ifr_rows[p, q][timeslice]
                if not len(idx):
                    continue
                # this is the plot track ID. Things that we average over are replaced by None
                track = (iplot, None if average_ddids else ddid, None if average_ifrs else (p, q))
                # update label attributes
                labelattrs['baseline'] = baseline = round(ifrset.baseline(p, q))
                baselines.add(baseline)
                labelattrs['ifr'] = ifrlabel = ifrset.ifr_label(p, q)
                #
                if plot_type is PT_FLAGTRACK:
                    d1 = numpy.ma.masked_array(plotfunc(flagcol[idx, ...], meanaxis))
                else:
                    d1 = plotfunc(dc[idx, ...])
                    if plotreduce and d1.ndim > meanaxis:
                        d1 = getattr(d1, plotreduce)(meanaxis)
                    d1.fill_value = 0
                # check that this data track  is not fully flagged
                if d1.mask.all():
                    continue
                active_ifrs.add((p, q))
                # get a plot collection object, or make a new one
                plotcoll = get_plot_collection(track, plot_type)
                if options.offset:
                    plotcoll.offset = options.offset
                # for flag-density plots, set the plot offset
                if plot_type is PT_FLAGTRACK:
                    if not options.offset:
                        plotcoll.offset = max(plotcoll.offset, d1.max() * 0.11 * options.offset_std)
                # add or update track
                count = plotcoll.get_track_count(track)
                # count !=0 indicates track is being updated (because we're averaging DDIDs or IFRs)
                if count:
                    d0 = plotcoll.get_track_data(track)
                    if d0.shape != d1.shape:
                        warnings.warn("Shape mismatch between averaged DDIDs or IFRs, will ignore mismatched data.")
                        continue
                    # add data to track. For means, update as (old*n + new)/(n+1), but if stddev
                    # reduction was in effect, update as sqrt((old**2)*n + new)/(n-1)
                    if 'std' in (plotreduce, datareduce):
                        d1 = numpy.sqrt(((d0 ** 2) * count + d1 ** 2) / (count + 1))
                    else:
                        d1 = (d0 * count + d1) / (count + 1)
                # insert track data
                labelattrs['mean'] = mean = d1.mean()
                labelattrs['stddev'] = std = d1.std()
                if plot_type is PT_CC:
                    label = labelattrs['ifr'] if options.label_ifr else ""
                else:
                    label = label_format % labelattrs
                if options.group_redundant:
                    plotcoll.add_track(track, d1, count=count + 1, mean=mean, stddev=std, label=ifrlabel,
                                       group="%dm" % baseline)
                else:
                    plotcoll.add_track(track, d1, count=count + 1, mean=mean, stddev=std, label=label)

    # deallocate datacubes
    datacols = flagcol = dc = None

    # make list of active IFRS (as p,q pairs), sorted by baseline length
    print("===> Found data for %d interferometers" % len(active_ifrs))
    if not average_ifrs:
        from past.builtins import cmp
        from functools import cmp_to_key
        keyranges[2] = sorted(active_ifrs,
                              key=cmp_to_key(lambda a, b: cmp((round(ifrset.baseline(*a)), a), (round(ifrset.baseline(*b)), b))))

    if not active_ifrs:
        print("===> Nothing to be plotted. Check your data selection.")
        sys.exit(0)

    # Now, work out the order in which we plot the tracks and PlotCollections.
    # this is determined by the options.page and options.stack order

    # Form up list of all possible track keys, in the order in which they are
    # to be plotted. Note that after this first loop, the elements of each key are
    # scrambled, because of the way we build up the list...
    track_order = [[]]
    elorder = {}
    keyels = {PLOT: 0, DDID: 1, IFR: 2}
    for i, what in enumerate(options.average + options.stack + options.page):
        elorder[what] = i
        track_order = [to + [key] for key in keyranges[keyels[what]] for to in track_order]
    # ...so unscamble them into proper order and turn them into tuples
    track_order = [(key[elorder[PLOT]], key[elorder[DDID]], key[elorder[IFR]]) for key in track_order]

    # figure out paper size
    landscape = options.papertype[0] == "/"
    papertype = options.papertype[1:] if landscape else options.papertype

    # now, split this up into PlotCollections and the tracks associated with each
    pc_tracks = []
    pc0 = None
    for key in track_order:
        pc = plotcolls[key]
        # start new collection
        if pc is not pc0:
            pc_tracks.append((pc, [key]))
            pc0 = pc
        # else key belongs to previous collection
        else:
            pc_tracks[-1][1].append(key)

    # Do we have multiple plots, or multiple pages per plot? Start a page counter then.
    if len(pc_tracks) > 1 or len(pc_tracks[0][1]) > options.ppp:
        ipage = 0
    else:
        ipage = None

    # loop over plotcollectors
    for iplotcoll, (plotcoll, keylist) in enumerate(pc_tracks):
        # do we have a title for this plot on the command line?
        if options.title and len(options.title) > iplotcoll:
            title0 = options.title[iplotcoll]
        # no, make default title
        else:
            nplot, ddid, ifr = keylist[0]
            chans = freqslice.indices(datashape[1])
            if chans[1] == chans[0] + 1:
                chandesc = str(chans[0])
            elif chans[2] == 1:
                chandesc = "%d~%d" % (chans[0], chans[1])
            else:
                chandesc = "%d~%d step %d" % (chans[0], chans[1] - 1, chans[2])
            title0 = msname
            if average_ddids:
                title0 += " ddid %s" % "+".join(map(str, ddids))
            else:
                title0 += " ddid %d" % ddid
            if options.freqslice:
                title0 += " (chan %s)" % chandesc
            if average_ifrs:
                title0 += " mean of %d ifrs" % len(active_ifrs)
            elif IFR in options.page:
                title0 += " ifr %s (%dm)" % (ifrset.ifr_label(*ifr), round(ifrset.baseline(*ifr)))
            # plot description
            if PLOT in options.page:
                title0 += " " + plots[nplot][1]
        # split keylist if needed
        if options.group_redundant:
            print("===> Grouping by %d unique baselines" % len(baselines))
            ppp = options.ppp
            baselines = sorted(baselines)
            for nkey0 in range(0, len(baselines), ppp):
                nkey1 = min(nkey0 + options.ppp, len(baselines))
                grouplist = ["%dm" % bl for bl in baselines[nkey0:nkey1]]
                # make filename, if saving file
                savefile = options.output
                if ipage is not None:
                    if savefile:
                        basename, ext = os.path.splitext(savefile)
                        savefile = "%s.%d%s" % (basename, ipage, ext)
                    ipage += 1
                dual = (len(grouplist) > options.ppp / 2)
                plotcoll.make_figure(grouplist, save=savefile, suptitle=title0, figsize=figsize,
                                     xgrid=options.x_grid or True, ygrid=not options.no_y_grid,
                                     papertype=papertype, dpi=options.resolution, dual=dual, landscape=landscape)
        else:
            ppp = options.ppp if type(plotcoll) is Owlcat.Plotting.PlotCollection else len(keylist)
            for nkey0 in range(0, len(keylist), ppp):
                nkey1 = min(nkey0 + options.ppp, len(keylist))
                keys = keylist[nkey0:nkey1]
                # make filename, if saving file
                savefile = options.output
                if ipage is not None:
                    if savefile:
                        basename, ext = os.path.splitext(savefile)
                        savefile = "%s.%d%s" % (basename, ipage, ext)
                    ipage += 1
                dual = (len(keys) > options.ppp / 2)
                plotcoll.make_figure(keys, save=savefile, suptitle=title0, figsize=figsize,
                                     offset_std=options.offset_std,
                                     xgrid=options.x_grid or True, ygrid=not options.no_y_grid,
                                     papertype=papertype, dpi=options.resolution, dual=dual, landscape=landscape)

    if not options.output:
        from pylab import plt

        plt.show()



# -*- coding: utf-8 -*-
import re


def parse_slice(spec, multiplier=1):
    """Parses a slice specification like "x" or x:y" or "x:y:step" or "x~y" or "x~y:step".
    The x:y form is an exclusive range, x~y is inclusive.
    Returns a slice object.
    """
    # null spec -- return "all" slice
    if not spec:
        return slice(0, None)
    match = re.match("^(?P<start>\d*)((?P<sep>[:~])(?P<end>\d*)(:(?P<step>\d*))?)?$", spec)
    if not match:
        raise ValueError("invalid slice specification %s" % spec)
    start, end, step = [((match.group(name) or None) and int(match.group(name))) for name in ('start', 'end', 'step')]
    sep = match.group('sep')
    # No match for first separator implies single number. If even that is missing, return full slice
    if sep == None:
        return slice(start * multiplier, (start + 1) * multiplier) if start is not None else slice(0, None)
    # increment end if specified inclusive slice as "start~end"
    elif sep == "~" and end is not None:
        end += 1
    return slice(start and start * multiplier, end and end * multiplier, step and step * multiplier)



# -*- coding: utf-8 -*-
import numpy as np
import re
import tempfile
import os
from past.builtins import cmp
from functools import cmp_to_key

import Owlcat
from Kittens.utils import verbosity
try:
    import Purr.Pipe

    has_purr = True
except:
    has_purr = False

_gli = None # Meow.MSUtils.find_exec('glish')


# Various argument-formatting methods to use with the Flagger.AutoFlagger class
# below. These really should be static methods of the class, but that doesn't work
# with Python (specifically, I cannot include them into static member dicts)
def _format_nbins(bins, argname):
    if isinstance(bins, (list, tuple)) and len(bins) == 2:
        return str(list(bins))
    else:
        return "%d" % bins
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, bins))


def _format_bool(arg, argname):
    if arg:
        return "T"
    else:
        return "F"


def _format_index(arg, argname):
    if isinstance(arg, int):
        if arg < 0:
            return str(arg)
        else:
            return str(arg + 1)
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))


def _format_list(arg, argname):
    if isinstance(arg, str):
        return "'%s'" % arg
    elif isinstance(arg, (list, tuple)):
        return "[%s]" % (','.join([_format_list(x, argname) for x in arg]))
    elif isinstance(arg, bool):
        return _format_bool(arg, argname)
    elif isinstance(arg, (int, float)):
        return str(arg)
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))


def _format_ilist(arg, argname):
    if isinstance(arg, str):
        return "'%s'" % arg
    elif isinstance(arg, (list, tuple)):
        return "[%s]" % (','.join([_format_ilist(x, argname) for x in arg]))
    elif isinstance(arg, bool):
        return _format_bool(arg, argname)
    elif isinstance(arg, int):
        return _format_index(arg, argname)
    elif isinstance(arg, float):
        return str(arg)
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))


def _format_plotchan(arg, argname):
    if isinstance(arg, bool):
        return _format_bool(arg, argname)
    else:
        return _format_index(arg, argname)


def _format_2N(arg, argname):
    if isinstance(arg, (list, tuple)):
        return "%s" % arg
    elif numpy.isarray(arg):
        if arg.dtype is np.int32:
            a = arg + 1
        else:
            a = arg
        if a.ndim == 1:
            return str(list(a))
        elif a.ndim == 2 and a.shape[1] == 2:
            return "[%s]" % (','.join([str(list(a[i, :])) for i in range(a.shape[0])]))
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))


def _format_clip(arg, argname):
    if isinstance(arg, (list, tuple)):
        if not isinstance(arg[0], (list, tuple)):
            arg = [arg]
        recfields = []
        for i, triplet in enumerate(arg):
            if not isinstance(triplet, (list, tuple)) or len(triplet) != 3:
                raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))
            expr, minval, maxval = triplet
            subfields = ["expr='%s'" % expr]
            if minval is not None:
                subfields.append("min=%g" % minval)
            if maxval is not None:
                subfields.append("max=%g" % maxval)
            recfields.append("a%d=[%s]" % (i, ','.join(subfields)))
        return "[%s]" % ','.join(recfields)
    raise TypeError("invalid value for '%s' keyword (%s)" % (argname, arg))


# mapping from casacore table valueType to bit length
BITFLAG_NBITS = dict(uchar=8, short=16, int=32)

# mapping from casacore table valueType to numpy dtype
BITFLAG_DTYPES = dict(uchar=np.uint8, short=np.int16, int=np.int32)



class Flagsets(object):
    """ Manage an measurement set's flagsets. Pasted from Cattery.Meow.MSUtils. """

    def __init__(self, ms):
        """
        Initialises a Flagsets object for the measurement set.

        Args:
            ms (:obj:`~casacore.tables.table.table`):
                A table object belonging to the measurement set.
        """

        self.ms = ms
        if not 'BITFLAG' in ms.colnames():
            self.order = None
            self.bits = {}
            self.NBITS = 8
            self.dtype = np.uint8
        else:
            value_type = ms.getcoldesc('BITFLAG')['valueType']
            self.NBITS = BITFLAG_NBITS[value_type]
            self.dtype = BITFLAG_DTYPES[value_type]
            kws = ms.colkeywordnames('BITFLAG')
            self.bits = {}
            # scan FLAGSET_xxx keywords and populate name->bitmask mappings
            for kw in kws:
                match = re.match('^FLAGSET_(.*)$', kw)
                if match:
                    name = match.group(1)
                    bit = ms.getcolkeyword('BITFLAG', kw)
                    if isinstance(bit, int):
                        self.bits[name] = bit
                    else:
                        print("Warning: unexpected type (%s) for %s keyword of BITFLAG column," \
                              " ignoring" % (type(bit), kw))
            # have we found any FLAGSET_ specs?
            if self.bits:
                order = 'FLAGSETS' in kws and ms.getcolkeyword('BITFLAG', 'FLAGSETS')
                if isinstance(order, str):
                    order = order.split(',')
                else:
                    print("Warning: unexpected type (%s) for FLAGSETS keyword of BITFLAG column," \
                          " ignoring" % type(order))
                    order = []
                # form up "natural" order by comparing bitmasks
                bitwise_order = list(self.bits.keys())
                bitwise_order.sort(key=cmp_to_key(lambda a, b: cmp(self.bits[a], self.bits[b])))
                # if an order is specified, make sure it is actually valid,
                # and add any elements from bitwise_order that are not present
                self.order = [fs for fs in order if fs in self.bits] + \
                             [fs for fs in bitwise_order if fs not in order]
                # if order was fixed compared to what was in MS, write back to MS
                if ms.iswritable() and self.order != order:
                    ms._putkeyword('BITFLAG', 'FLAGSETS', -1, False, ','.join(self.order))
                    ms.flush()
            # else if no flagsets found, try the old-style NAMES keyword
            elif 'NAMES' in kws:
                names = ms.getcolkeyword('BITFLAG', 'NAMES')
                if isinstance(names, (list, tuple)):
                    self.order = list(map(str, names))
                    bit = 1
                    for name in self.order:
                        self.bits[name] = bit
                        bit <<= 1
                    if ms.iswritable():
                        ms._putkeyword('BITFLAG', 'FLAGSETS', -1, False, ','.join(self.order))
                        for name, bit in self.bits.items():
                            ms._putkeyword('BITFLAG', 'FLAGSET_%s' % name, -1, False, bit)
                        ms.flush()
            else:
                self.order = []
        self.LEGACY = 1 << self.NBITS
        self.BITMASK_ALL = self.LEGACY - 1

    def names(self):
        """
        Convenience function for determining active flagsets.

        Returns:
            list or None:
                A list of flagset names, in the order in which they were created or None if BITFLAG
                column is missing (so flagsets are unavailable.)
        """

        return self.order

    def flagmask(self, name, create=False):
        """
        Flagmask corresponding to named flagset.

        Args:
            name (str):
                Name of flagset.
            create (bool, optional):
                If True and flagset is missing, creates named flagset, else raises exception.

        Raises:
            TypeError:
                If the MS does not contain a BITFLAG column.
            ValueError:
                If the named flagset is not found and create is False.
            ValueError:
                If there are too many flagsets to create a new one.
        """

        # Cludge for python2/3 interoperability.
        name = str(name)

        # lookup flagbit, return if found
        if self.order is None:
            raise TypeError("MS does not contain a BITFLAG column. Please run the addbitflagcol" \
                            " utility on this MS.")
        bit = self.bits.get(name, None)
        if bit is not None:
            return bit
        # raise exception if not allowed to create a new one
        if not create:
            raise ValueError("Flagset '%s' not found" % name)
        # find empty bit
        for bitnum in range(self.NBITS):
            bit = 1 << bitnum
            if bit not in list(self.bits.values()):
                self.order.append(name)
                self.bits[name] = bit
                self.ms._putkeyword('BITFLAG', 'FLAGSETS', -1, False, ','.join(self.order))
                self.ms._putkeyword('BITFLAG', 'FLAGSET_%s' % name, -1, False, bit)
                self.ms.flush()
                return bit
        # no free bit found, bummer
        raise ValueError("Too many flagsets in MS, cannot create another one")

    def remove_flagset(self, *fsnames):
        """
        Removes the named flagset(s).

        Args:
            fsnames (tuple):
                Names of flagsets to be removed.

        Returns:
            int:
                Flagmask corresponding to the removed flagsets.
        """

        # lookup all flagsets, raise error if any not found
        if self.bits is None:
            raise TypeError("MS does not contain a BITFLAG column, cannot use flagsets")
        removing = []
        for fs in fsnames:
            bit = self.bits.get(fs, None)
            if bit is None:
                raise ValueError("Flagset '%s' not found" % fs)
            removing.append((fs, bit))
        if not removing:
            return
        # remove items, form up mask of bitflags to be cleared
        mask = 0
        for name, bit in removing:
            mask |= bit
            del self.bits[name]
            del self.order[self.order.index(name)]
            self.ms.removecolkeyword('BITFLAG', 'FLAGSET_%s' % name)
        # write new list of bitflags
        self.ms._putkeyword('BITFLAG', 'FLAGSETS', -1, False, ','.join(self.order))
        self.ms.flush()

        return mask


class Flagger(verbosity):
    def __init__(self, msname, verbose=0, timestamps=False, chunksize=200000):
        verbosity.__init__(self, name="Flagger")
        self.set_verbose(verbose)
        if timestamps:
            self.enable_timestamps(modulo=10000)
        self.msname = msname
        self.ms = None
        self.readwrite = False
        self.chunksize = chunksize
        self._initializing_bitflags = False  # will be true if initializing
        self._reopen()

    def close(self):
        if self.ms:
            self.dprint(2, "closing MS", self.msname)
            self.ms.close()
        self.ms = self.flagsets = None

    def _reopen(self, readwrite=False):
        if self.ms is None:
            self.ms = ms = Owlcat.table(self.msname, readonly=not readwrite, ack=False)
            self.readwrite = readwrite
            self.dprintf(1, "opened MS %s, %d rows\n", ms.name(), ms.nrows())
            has_bf = 'BITFLAG' in ms.colnames()
            has_bfr = 'BITFLAG_ROW' in ms.colnames()
            self.has_bitflags = has_bf and has_bfr
            if self.has_bitflags:
                self.has_bitflags = self.ms.getcoldesc("BITFLAG")['valueType']
                self.dprint(1, "this MS has proper bitflag columns of type '%s'" % self.has_bitflags)
            else:
                if has_bf or has_bfr:
                    self.dprint(0, "WARNING: bitflag columns only partially initialized. Try --reinit-bitflags?")
            self.flagsets = Flagsets(ms)
            self.dprintf(1, "flagsets are %s\n", self.flagsets.names())
            self.purrpipe = Purr.Pipe.Pipe(self.msname) if has_purr else None
            self.LEGACY = self.flagsets.LEGACY
            self.BITMASK_ALL = self.flagsets.BITMASK_ALL

        elif self.readwrite != readwrite:
            self.ms = Owlcat.table(self.msname, readonly=not readwrite, ack=False)
            self.readwrite = readwrite

        return self.ms

    def _add_column(self, col_name, like_col="DATA", like_type=None, stman=None):
        """
        Inserts a new column into the measurement set.

        Args:
            col_name (str):
                Name of target column.
            like_col (str, optional):
                Column will be patterned on the named column.
            like_type (str or None, optional):
                If set, column type will be changed.

        Returns:
            bool:
                True if a new column was inserted, else False.
        """

        if col_name in self.ms.colnames():
            return False
        self._reopen(True)
        # new column needs to be inserted -- get column description from column 'like_col'
        desc = self.ms.getcoldesc(like_col)
        desc['name'] = col_name
        desc['comment'] = desc['comment'].replace(" ", "_")  # casacore not fond of spaces...
        dminfo = self.ms.getdminfo(like_col)
        dminfo["NAME"] =  "{}-{}".format(dminfo["NAME"], col_name)
        if stman is not None:
            dminfo["TYPE"] = stman
        # if a different type is specified, insert that
        if like_type:
            desc['valueType'] = like_type
        self.dprint(0, "inserting new column %s of type '%s' based on %s" % (col_name, desc['valueType'], like_col))
        if stman is not None:
            self.dprint(0, "  using %s" % stman)
        self.ms.addcols(desc, dminfo)
        return True

    def remove_bitflags(self):
        removecols = [col for col in ("BITFLAG", "BITFLAG_ROW") if col in self.ms.colnames()]
        if removecols:
            self._reopen(True)
            self.dprint(0, "removing columns", ", ".join(removecols))
            self.ms.removecols(removecols)
            self.close()
            self._reopen()

    def add_bitflags(self, wait=True, purr=True, bits=32, stman=None):
        if not self.has_bitflags:
            self._reopen(True)
            if bits == 32:
                dtype = 'int'
                self.has_bitflags = np.int32
            elif bits == 16:
                dtype = 'short'
                self.has_bitflags = np.int16
            elif bits == 8:
                dtype = 'uchar'
                self.has_bitflags = np.uint16
            else:
                raise ValueError("invalid bits setting. 8, 16 or 32 expected.")
            self._add_column("BITFLAG", "DATA", dtype, stman=stman)
            self._add_column("BITFLAG_ROW", "FLAG_ROW", dtype, stman=stman)
            self.close()
            self._reopen()
            self._initializing_bitflags = True

    def remove_flagset(self, *fsnames, **kw):
        self._reopen(True)
        mask = self.flagsets.remove_flagset(*fsnames)
        # report to purr
        if kw.get('purr', True) and self.purrpipe:
            self.purrpipe.title("Flagging").comment("Removing flagset(s) %s." % (','.join(fsnames)))
        self.unflag(mask, purr=False)


    def _get_submss(self, ms, ddids=None):
        """Helper method. Splits MS into subsets by DATA_DESC_ID.
        Returns list of (ddid,nrows,subms) tuples, where subms is a subset of the MS with the given DDID,
        and nrows is the number of rows in preceding submss (which is needed for progress stats)
        """
        # build up list of (ddid,nrows,subms) pairs.
        # subms is a subset of the MS with the given DDID
        # nrows is the number of rows in preceding submss (need for progress stats)
        sub_mss = []
        nrows = 0
        if ddids is None:
            ddids = list(range(TABLE(ms.getkeyword('DATA_DESCRIPTION'), ack=False).nrows()))
        for ddid in ddids:
            subms = ms.query("DATA_DESC_ID==%d" % ddid)
            sub_mss.append((ddid, nrows, subms))
            nrows += subms.nrows()
        return sub_mss


    def lookup_flagmask(self, flagset, create=False):
        """helper function: converts a flagset name into an integer flagmask"""
        if flagset is None:
            return None
        # convert flagset to list
        elif isinstance(flagset, int):
            return int(flagset)
        elif isinstance(flagset, str):
            flagset = flagset.split(",")
        elif not isinstance(flagset, (list, tuple)):
            raise TypeError("invalid flagset of type %s" % type(flagset))
        # loop over list and accumulate flagmask
        flagmask = 0
        for fset in flagset:
            if isinstance(fset, int):
                flagmask |= fset
            elif isinstance(fset, str):
                if fset == "ALL":
                    flagmask |= self.BITMASK_ALL | self.LEGACY
                elif fset.upper() == "ALL":
                    flagmask |= self.BITMASK_ALL
                else:
                    # +L suffix includes legacy flags
                    if fset[-2:].upper() == "+L":
                        flagmask |= self.LEGACY
                        fset = fset[:-2]
                    # lookup name or number
                    if re.match('^\d+$', fset):
                        flagmask |= int(fset)
                    elif re.match('^0[xX][\dA-Fa-f]+$', fset):
                        flagmask |= int(fset, 16)
                    elif fset:
                        flagmask |= self.flagsets.flagmask(fset, create=create)
            else:
                raise TypeError("invalid flagset of type %s in list of flagsets" % type(fset))
        return flagmask

    def flagmaskstr(self, flagmask):
        """helper function: converts an integer flagmask into a printable str"""
        if not flagmask:
            return "0"
        legacy = flagmask & self.LEGACY
        bits = flagmask & self.BITMASK_ALL
        return "0x%04X%s" % (bits, "+L" if legacy else "")

    def xflag(self,
              # These options determine what to do with the subset determined below. If none are set,
              # the xflag() simply returns stats without doing anything
              flag=None,  # set the flagmask (or flagset name, optionally "+L")
              unflag=None,  # clear the flagmask (or flagset name, optionally "+L")
              copy=None,    # copy to flagmask (or flagset name, optionally "+L")
              create=False,  # if True and 'flag' is a string, creates new flagset as needed
              fill_legacy=None,  # if not None, legacy flags will be filled (within all of subset A) using this mask

              # The following options restrict the subset to be flagged. Effect is cumulative.
              # Row subset. Data selection by whole rows
              ddid=None, fieldid=None,  # DATA_DESC_ID or FIELD_ID, single int or list
              antennas=None,  # list of antennas
              baselines=None,  # list of baselines (as ip,iq pairs)
              time=None,  # time range as (min,max) pair (can use None for either min or max)
              reltime=None,  # relative time (rel. to start of MS) range as (min,max) pair
              taql=None,  # additional TaQL string
              # Subset A. Freq/corr slices within the row subset.
              channels=None,  # channel subset (index, or slice, or list of index/slices)
              corrs=None,  # correlation subset (index, or slice, or list of index/slices)
              # Subset B. Subset within subset A based on flags
              flagmask=None,  # any bitflag set in the given flagmask (or flagset name)
              flagmask_all=None,  # all bitflags set in the given flagmask (or flagset name)
              flagmask_none=None,  # no bitflags set in the given flagmask (or flagset name)
              # Subset C. Subset within subset B based on data clipping
              data_nan=False,  # restrict flagged subset to NAN/INF data
              data_above=None,  # restrict flagged subset to abs(data)>X
              data_below=None,  # and abs(data)<X
              data_fm_above=None,  # same as data_above/_below, but flags based on the mean
              data_fm_below=None,  # amplitude across all frequencies
              data_column='CORRECTED_DATA',  # data column for clip_above and clip_below
              data_flagmask=-1,  # flagmask to apply to data column when computing mean
              # if True, dict of {name: mask} bitflags for which statistics will be collected
              get_stats_only=None,
              # other options
              flag_allcorr=True,  # flag all correlations if at least one is flagged
              progress_callback=None,  # callback, called with (n,nmax) to report progress
              purr=False  # if True, writes comments to purrpipe
              ):
        """Alternative flag interface, works on the in/out principle."""
        if not self.purrpipe:
            purr = False
        ms = self._reopen(flag or unflag or copy or fill_legacy is not None)
        # lookup flagset names
        for var in 'flag', 'unflag', 'copy', 'flagmask', 'flagmask_all', 'flagmask_none', 'data_flagmask', 'fill_legacy':
            flagval = locals()[var]
            locals()[var] = fm = self.lookup_flagmask(flagval, create=(var == 'flag'))
            if flagval is not None:
                self.dprintf(2, "%s=%s corresponds to bitmask %s\n", var, flagval, self.flagmaskstr(fm))
        # for these two masks, it's more convenient that they're set to 0 if missing
        flag = flag or 0
        unflag = unflag or 0
        copy = copy or 0
        if not self.has_bitflags and (flag | unflag | copy) & self.BITMASK_ALL:
            raise RuntimeError("no BITFLAG column in this MS, can't change bitflags")
        # stats accumulator
        if get_stats_only:
            stats = { name: 0 for name in get_stats_only.keys() }
        # total number of rows
        totrows = ms.nrows()
        # rows and visibilities in the row subset
        sel_nrow = sel_nvis = 0
        # visibilities selected in subsets A, B and C
        nvis_A = nvis_B = nvis_C = 0
        # get DDIDs and FIELD_IDs
        if ddid is None:
            ddids = list(range(Owlcat.table(ms.getkeyword('DATA_DESCRIPTION'), ack=False, readonly=True).nrows()))
        elif isinstance(ddid, int):
            ddids = [ddid]
        elif isinstance(ddid, (tuple, list)):
            ddids = ddid
        else:
            raise TypeError("invalid ddid argument of type %s" % type(ddid))
        # form up list of TaQL expressions for subset selectors
        queries = []
        if taql:
            queries.append(taql)
        if fieldid is not None:
            if isinstance(fieldid, int):
                fieldid = [fieldid]
            elif not isinstance(fieldid, (tuple, list)):
                raise TypeError("invalid fieldid argument of type %s" % type(fieldid))
            queries.append(" || ".join(["FIELD_ID==%d" % f for f in fieldid]))
        if antennas is not None:
            antlist = str(list(antennas))
            queries.append("ANTENNA1 in %s || ANTENNA2 in %s" % (antlist, antlist))
        if time is not None:
            t0, t1 = time
            if t0 is not None:
                queries.append("TIME>=%g" % t0)
            if t1 is not None:
                queries.append("TIME<=%g" % t1)
        if reltime is not None:
            t0, t1 = reltime
            time0 = self.ms.getcol('TIME', 0, 1)[0]
            if t0 is not None:
                queries.append("TIME>=%f" % (time0 + t0))
            if t1 is not None:
                queries.append("TIME<=%f" % (time0 + t1))
        # form up TaQL string, and extract subset of table
        if queries:
            query = "( " + " ) && ( ".join(queries) + " )"
            purr and self.purrpipe.comment("; effective MS selection is \"%s\"" % query, endline=False)
            self.dprintf(2, "selection string is %s\n", query)
            ms = ms.query(query)
            self.dprintf(2, "query reduces MS to %d rows\n", ms.nrows())
        else:
            self.dprintf(2, "no selection applied\n")
        # check list of baselines
        if baselines:
            baselines = [(int(p), int(q)) for p, q in baselines]
            purr and self.purrpipe.comment("; baseline subset is %s" %
                                           " ".join(["%d-%d" % (p, q) for p, q in baselines]),
                                           endline=False)

        # helper func to parse the channels/corrs/timeslots arguments
        def make_slice_list(selection, parm):
            if not selection:
                return [np.s_[:]]
            if isinstance(selection, (int, slice)):
                return make_slice_list([selection], parm)
            if not isinstance(selection, (list, tuple)):
                raise TypeError("invalid %s selection: %s" % (parm, selection))
            sellist = []
            for sel in selection:
                if isinstance(sel, int):
                    sellist.append(slice(sel, sel + 1))
                elif isinstance(sel, slice):
                    sellist.append(sel)
                else:
                    raise TypeError("invalid %s selection: %s" % (parm, selection))
            return sellist

        # parse the arguments
        channels = make_slice_list(channels, 'channels')
        corrs = make_slice_list(corrs, 'corrs')
        purr and self.purrpipe.comment("; channels are %s" % channels, endline=False)
        purr and self.purrpipe.comment("; correlations are %s" % corrs, endline=False)
        # put comment into purrpipe
        purr and self.purrpipe.comment(".")
        #
        flagsubsets = flagmask is not None or flagmask_all is not None or flagmask_none is not None
        dataclip = data_above is not None or data_below is not None or data_nan or \
                   data_fm_above is not None or data_fm_below is not None
        # make list of sub-MSs by DDID
        sub_mss = self._get_submss(ms, ddids)
        nrow_tot = ms.nrows()
        # go through rows of the MS in chunks
        for ddid, irow_prev, ms in sub_mss:
            self.dprintf(2, "processing MS subset for ddid %d\n", ddid)
            if progress_callback:
                progress_callback(irow_prev, nrow_tot)
            for row0 in range(0, ms.nrows(), self.chunksize):
                if progress_callback:
                    progress_callback(irow_prev + row0, nrow_tot)
                nrows = min(self.chunksize, ms.nrows() - row0)
                self.dprintf(2, "processing rows %d:%d (%d rows total)\n", row0, row0 + nrows - 1, nrows)
                # apply baseline selection to the mask
                if baselines:
                    # rowmask will be True for all selected rows
                    rowmask = np.zeros(nrows, bool)
                    a1 = ms.getcol('ANTENNA1', row0, nrows)
                    a2 = ms.getcol('ANTENNA2', row0, nrows)
                    # update mask
                    for p, q in baselines:
                        rowmask |= (a1 == p) & (a2 == q)
                    self.dprintf(2, "baseline selection leaves %d rows\n", nr)
                # else select all rows
                else:
                    # rowmask will be True for all selected rows
                    rowmask = np.ones(nrows, bool)
                # read legacy flags to get a datashape
                legacy_flag_column = ms.getcol('FLAG', row0, nrows)
                legacy_flag_column |= ms.getcol('FLAG_ROW', row0, nrows)[:, np.newaxis, np.newaxis]
                datashape = legacy_flag_column.shape
                nv_per_row = datashape[1] * datashape[2]
                # rowflags and visflags will be constructed on-demand below. Make helper functions for this
                self._visflags = None

                def bitflags():
                    """Reads BITFLAG column on demand. If no such column, forms up zero array"""
                    if self._visflags is None:
                        if not self.has_bitflags or self._initializing_bitflags:
                            self._visflags = np.zeros_like(legacy_flag_column, self.flagsets.dtype)
                            self.dprint(2, "formed empty BITFLAGs (type {})".format(self.flagsets.dtype))
                        elif all([ms.iscelldefined('BITFLAG', r) for r in range(row0, row0+nrows)]):
                            self._visflags = ms.getcol('BITFLAG', row0, nrows)
                            self.dprint(2, "read BITFLAG column")
                            self._visflags |= ms.getcol('BITFLAG_ROW', row0, nrows)[:, np.newaxis, np.newaxis]
                            self.bf_type = self._visflags.dtype.type
                        else:
                            self._visflags = np.zeros_like(legacy_flag_column, self.flagsets.dtype)
                            self.dprint(2, "formed empty BITFLAGs (type {}) because part (or all) of the column is undefined".format(self.flagsets.dtype))
                        self._bf_type = self._visflags.dtype.type
                    return self._visflags

                # apply stats
                nr = rowmask.sum()
                sel_nrow += nr
                nv = nr * nv_per_row
                sel_nvis += nv
                self.dprintf(2, "Row subset (data selection) leaves %d rows and %d visibilities\n", nr, nv)
                # get subset C
                # vismask will be True for all selected visibilities
                vismask = np.zeros(datashape, bool)
                for channel_slice in channels:
                    for corr_slice in corrs:
                        vismask[rowmask, channel_slice, corr_slice] = True
                self.dprintf(3, "  finished applying freq/corr slicing\n")
                nv = vismask.sum()
                nvis_A += nv                
                self.dprintf(2, "subset A (freq/corr slicing) leaves %d visibilities\n", nv)
                # stats-only mode
                if get_stats_only:
                    for name, mask in get_stats_only.items():
                        if mask is None:
                            nv = np.count_nonzero(legacy_flag_column)
                        else:
                            vf1 = bitflags()&mask
                            nv = np.count_nonzero(vf1)
                        stats[name] += nv
                        self.dprintf(3, "flagset %s flags %d visibilities\n", name, nv)
                    continue  # to top of loop to get next chunk

                # select based on flag state
                if flagsubsets:
                    # select using the "any" scheme
                    if flagmask is not None:
                        if flagmask == self.LEGACY:
                            vismask &= legacy_flag_column
                        else:
                            bf1 = (bitflags()&flagmask) != 0
                            if flagmask&self.LEGACY:
                                bf1 |= legacy_flag_column
                            vismask &= bf1
                    # select using the "all" scheme
                    if flagmask_all is not None:
                        if flagmask_all == self.LEGACY:
                            vismask &= legacy_flag_column
                        else:
                            fm = flagmask_all & ~self.LEGACY
                            bf1 = (bitflags()&fm) == fm
                            if flagmask_all&self.LEGACY:
                                bf1 &= legacy_flag_column
                            vismask &= bf1
                    # select using the "None" scheme
                    if flagmask_none is not None:
                        if flagmask_none == self.LEGACY:
                            vismask &= ~legacy_flag_column
                        else:
                            fm = flagmask_none & ~self.LEGACY
                            bf1 = (bitflags()&flagmask_none) == 0
                            if flagmask_none&self.LEGACY:
                                bf1 &= ~legacy_flag_column
                            vismask &= bf1
                nv = vismask.sum()
                nvis_B += nv
                self.dprintf(2, "subset B (flag-based selection) leaves %d visibilities\n", nv)
                # now apply clipping
                if dataclip:
                    datacol = ms.getcol(data_column, row0, nrows)
                    # make it a masked array: mask out stuff not in vismask
                    datamask = ~vismask
                    # and mask stuff in data_flagmask
                    if data_flagmask is not None:
                        datamask |= ((bitflags() & data_flagmask) != 0)
                    datacol = np.ma.masked_array(datacol, datamask)
                    self.dprintf(4, "datamask contains %d masked visibilities\n", datamask.sum())
                    self.dprintf(3, "At start of clipping we have %d visibilities\n", vismask.sum())
                    self.dprintf(3, "of which %d are finite\n", np.isfinite(datacol).sum())
                    # clip on NANs
                    if data_nan:
                        vismask &= ~np.isfinite(datacol)
                        self.dprintf(3, "NAN filtering leaves %d visibilities\n", vismask.sum())
                    # clip on amplitudes
                    abscol = abs(datacol)
                    if data_above is not None:
                        vismask &= abscol > data_above
                        self.dprintf(3, "data_above filtering leaves %d visibilities\n", vismask.sum())
                    if data_below is not None:
                        vismask &= abscol < data_below
                        self.dprintf(3, "data_below filtering leaves %d visibilities\n", vismask.sum())
                    # clip on freq-mean amplitudes
                    if data_fm_above is not None or data_fm_below is not None:
                        datacol = abscol.mean(1)
                        #            print datacol.max(),data_fm_above
                        if data_fm_above is not None:
                            vismask &= (datacol > data_fm_above)[:, np.newaxis, ...]
                        if data_fm_below is not None:
                            vismask &= (datacol < data_fm_below)[:, np.newaxis, ...]
                        self.dprintf(3, "data_fm_above/below filtering leaves %d visibilities\n", vismask.sum())
                # finally, subset E is ready
                nv = vismask.sum()
                self.dprintf(2, "subset C (data clipping) leaves %d visibilities\n", nv)
                # extending flagging to all correlations
                if flag_allcorr:
                    vismask |= np.logical_or.reduce(vismask, 2)[:, :, np.newaxis]
                    nv = vismask.sum()
                    self.dprintf(2, "which extends to %d visibilities with flag_allcorr in effect\n", nv)
                nvis_C += nv

                # ok, at this point vismask is a boolean array of things to flag or unflag.
                # BITFLAGS would only have been read in if necessary

                # now, do the actual flagging
                if flag or unflag or copy or fill_legacy is not None:
                    changing_legacy_flags = fill_legacy is not None or (flag|unflag|copy)&self.LEGACY
                    # flag/unflag visibilities
                    if flag:
                        self.dprint(4, "doing flag")
                        if flag&self.LEGACY:
                            legacy_flag_column[vismask] = True
                        if flag&~self.LEGACY:
                            bitflags()[vismask] |= self._bf_type(flag&~self.LEGACY)
                    if copy:
                        self.dprint(4, "doing flag-copy")
                        if copy&self.LEGACY:
                            legacy_flag_column = vismask
                        if copy&~self.LEGACY:
                            bf = bitflags()
                            bf[vismask] |= self._bf_type(copy&~self.LEGACY)
                    if unflag:
                        self.dprint(4, "doing unflag")
                        if unflag&self.LEGACY:
                            legacy_flag_column[unflag] = False
                        if unflag&~self.LEGACY:
                            bitflags()[vismask] &= ~self._bf_type(unflag&~self.LEGACY)
                    # fill legacy flags
                    if fill_legacy is not None:
                        self.dprint(4, "filling legacy flags")
                        bf = bitflags()
                        legacy_flag_column[rowmask,...] = (bf[rowmask,...] & self._bf_type(fill_legacy)) != 0
                    # adjust the rowflags
                    self.dprint(4, "adjusting rowflags")

                    # were any bitflag manipulations done -- write bitflags
                    if self.has_bitflags and (flag | unflag | copy) & ~self.LEGACY:
                        bf_row = np.bitwise_and.reduce(bitflags(), axis=(1,2))
                        ms.putcol('BITFLAG', bitflags(), row0, nrows)
                        ms.putcol('BITFLAG_ROW', bf_row, row0, nrows)
                        self.dprint(4, "wrote BITFLAG")
                    # write legacy flags
                    if changing_legacy_flags:
                        self.dprintf(4, "filling legacy flags for rows %d:%d\n" % (row0, row0 + nrows))
                        ms.putcol('FLAG', legacy_flag_column, row0, nrows)
                        ms.putcol('FLAG_ROW', legacy_flag_column.all(axis=(1, 2)), row0, nrows)
                        self.dprint(4, "wrote FLAG")
                self.dprint(4, "done with this chunk")
        if progress_callback:
            progress_callback(99, 100)
        self._rowflags = self._visflags = None
        # print collected stats
        self.dprint(1, "xflag stats:")
        self.dprintf(1, "total MS size:           %8d rows\n", totrows)
        self.dprintf(1, "data selection leaves    %8d rows, %8d visibilities\n", sel_nrow, sel_nvis)
        self.dprintf(1, "chan/corr slicing leaves %8s       %8d visibilities\n", '', nvis_A)
        self.dprintf(1, "flag selection leaves    %8s       %8d visibilities\n", '', nvis_B)
        self.dprintf(1, "data clipping leaves     %8s       %8d visibilities\n", '', nvis_C)

        if get_stats_only:
            for name, nv in stats.items():
                get_stats_only[name] = nv

        return totrows, sel_nrow, sel_nvis, nvis_A, nvis_B, nvis_C




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# % $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import time

time_start = time.time()

from Owlcat import table, tablecopy, tableexists, tabledelete, addImagingColumns
import Owlcat.Console

progress = Owlcat.Console.Reporter(timestamp=True)

SKIP_COLUMNS = set(("FLAG_CATEGORY", "NFRA_AVERAGEDATA", "PEELED_DATA", "UVMODEL_DATA"))

if __name__ == "__main__":

    # setup some standard command-line option parsing
    #
    from optparse import OptionParser

    parser = OptionParser(usage="""%prog: [options] MS_output MS_in1 MS_in2 ...""",
                          description="Concatenates several MSs into one. This is done on a row-by-row basis "
                                      "with no sanity checking, so it's up to you to insure that the input MSs actually "
                                      "belong together. All subtables and keywords of the output MS are inherited from "
                                      "the first input MS, unless the --renumber-spws option is in effect, in which case "
                                      "the SPECTRAL_WINDOW and DATA_DESCRIPTION tables are also merged (and spwids and "
                                      "ddids renumbered.)")
    parser.add_option("-a", "--append", action="store_true",
                      help="append input MSs to output MS")
    parser.add_option("-f", "--force", dest="force", action="store_true",
                      help="proceed without confirmation, and overwrite output MS if it already exists")
    parser.add_option("-s", "--renumber-spws", action="store_true",
                      help="treat each MS as a separate spectral window")
    parser.add_option("--renumber-obs", action="store_true",
                      help="renumber OBSERVATION_ID of each MS")
    parser.add_option("-v", "--verbose", action="store_true",
                      help="be more verbose")

    (options, msnames) = parser.parse_args()

    if len(msnames) < (2 if options.append else 3):
        parser.error("Insufficient number of arguments. Use '-h' for help.")

    msout = msnames[0]
    msins = msnames[1:]

    if tableexists(msout) and not options.append and not options.force:
        print(("Output MS", msout, "already exists. Use the -f switch to overwrite."))
        sys.exit(1)

    if not options.force:
        if options.append:
            print(("Will extend merged MS %s with %d input MSs:" % (msout, len(msins))))
        else:
            print(("Will create merged MS %s from %d input MSs:" % (msout, len(msins))))
        for ms in msins:
            print(("  ", ms))
        if options.renumber_spws:
            print("Each input MS will be put into a separate spectral window, spws will be renumbered.")
        else:
            print("Spectral windows will not be renumbered.")
        if options.renumber_obs:
            print("OBSERVATION_ID will be incremented for each MS")
        if input("Proceed (y/n)? ").strip().upper()[0] != "Y":
            print("Aborted by user.")
            sys.exit(1)

    obsid = 0

    if not options.append:
        if tableexists(msout) and options.force:
            progress("Deleting existing copy of %s" % msout)
            tabledelete(msout)
        # copy first MS to output as-is
        progress("Copying %s to %s" % (msins[0], msout))
        tablecopy(msins[0], msout, deep=True)
        # when renumbering spectral windows, need to have an addImagingColumns() call
        # unfortunately, this wipes CORRECTED_DATA! I see no way around this (copynorows=True is no help,
        # as then it fails to copy subtables) apart from the ugly: copy first MS deeply,
        # add imaging columns, then re-copy columns
        if options.renumber_spws:
            addImagingColumns(msout)
        # otherwise, first MS is already copied, so remove it from list
        # (but reset OBSERVATION_ID)
        else:
            if options.renumber_obs:
                tab = table(msout, readonly=False)
                obs = tab.getcol("OBSERVATION_ID")
                obs[:] = obsid
                obsid += 1
                tab.putcol("OBSERVATION_ID", obs)
                tab.close()
            msins = msins[1:]

    # open output for writing
    tab0 = table(msout, readonly=False)

    # get the number of DDIDs and SPWIDs
    ddid_offsets = [0] * len(msins)
    if options.renumber_spws:
        ddid_tab0 = table(tab0.getkeyword("DATA_DESCRIPTION"), readonly=False)
        spw_tab0 = table(tab0.getkeyword("SPECTRAL_WINDOW"), readonly=False)
        # now, somehow lwimager will still try to re-init MODEL_DATA and CORRECTED_DATA when dealing with merged
        # MSs. The only way around it that I can see is to fill up the ddid and spw tabs first
        progress("Filling new DATA_DESCRIPTION and SPECTRAL_WINDOW table")
        for num_ms, ms in enumerate(msins[1:]):
            tab = table(ms)
            ddid_tab = table(tab.getkeyword("DATA_DESCRIPTION"), readonly=False)
            nr0 = ddid_tab0.nrows()
            ddid_offsets[num_ms + 1] = nr0
            ddid_tab0.addrows(ddid_tab.nrows())
            for col in ddid_tab.colnames():
                data = ddid_tab.getcol(col)
                if col == "SPECTRAL_WINDOW_ID":
                    data += spw_tab0.nrows()
                ddid_tab0.putcol(col, data, nr0)
            # append content of the SPECTRAL_WINDOW
            spw_tab = table(tab.getkeyword("SPECTRAL_WINDOW"), readonly=False)
            nr0 = spw_tab0.nrows()
            spw_tab0.addrows(spw_tab.nrows())
            for col in spw_tab.colnames():
                data = spw_tab.getcol(col)
                spw_tab0.putcol(col, data, nr0)
        num_spws, num_ddids = spw_tab0.nrows(), ddid_tab0.nrows()
        ddid_tab0.close()
        spw_tab0.close()
        tab0.close()
        addImagingColumns(msout)
        tab0 = table(msout, readonly=False)

    overprint = progress if options.verbose else progress.overprint

    for num_ms, msname in enumerate(msins):
        tab = table(msname)
        nr0 = tab0.nrows()
        # in spw renumbering mode, because of the ugliness explained above,
        # we need to re-copy the data from the first MS
        if options.renumber_spws and not num_ms:
            nr0 = 0
        # else, insert more rows to accommodate added MS
        else:
            tab0.addrows(tab.nrows())
        progress("Current row count: %d" % nr0)
        progress("Merging in %d rows from %s" % (tab.nrows(), msname))
        for col in tab.colnames():
            if col not in SKIP_COLUMNS:
                overprint("Reading column %s" % col)
                try:
                    data = tab.getcol(col)
                except:
                    print(("WARNING: MS %s does not appear to contain a valid %s column" % (msname, col)))
                    print("   This is not necessarily fatal, proceeding with the merge anyway.")
                    continue
                if options.renumber_obs and col == "OBSERVATION_ID":
                    data[:] = obsid
                    obsid += 1
                # if renumbering DDIDs, increment the DDID column by the # of rows in the DDID table --
                # this ensures uniqueness of DDIDs
                if options.renumber_spws and col == "DATA_DESC_ID":
                    progress("Adding %d to DATA_DESC_ID" % ddid_offsets[num_ms])
                    data += ddid_offsets[num_ms]
                overprint("Writing column %s, shape %s" % (col, data.shape))
                tab0.putcol(col, data, nr0)

    overprint("Closing output MS %s\n" % msout)
    nr0 = tab0.nrows()
    progress("Wrote %d rows to output MS" % nr0)
    if options.renumber_spws:
        progress("Output MS contains %d spectral windows and %d DDIDs" % (num_spws, num_ddids))
    tab0.close()
